#!/usr/bin/python
# tray_icon.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (tray_icon.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: tray_icon.

Uses wxPython to show tray icon for BitDust. This is working inside
``bpmain`` process, uses wxreactor to connect with main Twisted loop.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import platform

#------------------------------------------------------------------------------

USE_TRAY_ICON = True
LABEL = 'BitDust'

_IconObject = None
_ControlFunc = None
_CurrentIcon = 'off'

_IconsDict = {}

_LinuxIcons = {
    'off': 'off24x24.png',
    'connected': 'connected24x24.png',
    'sync': 'sync24x24.png',
    'error': 'error24x24.png',
    'updated': 'updated24x24.png',
}

_WindowsIcons = {
    'off': 'off16x16.png',
    'connected': 'connected16x16.png',
    'sync': 'sync16x16.png',
    'error': 'error16x16.png',
    'updated': 'updated16x16.png',
}

_PopUpIconsDict = {
    'open': 'expand24x24.png',
    'sync': 'synchronize24x24.png',
    'restart': 'restart24x24.png',
    'reconnect': 'network24x24.png',
    'shutdown': 'shutdown24x24.png',
}

#------------------------------------------------------------------------------


def icons_dict():
    global _IconsDict
    return _IconsDict


def popup_icons_dict():
    global _PopUpIconsDict
    return _PopUpIconsDict


#------------------------------------------------------------------------------


def shutdown():
    global _IconObject
    global USE_TRAY_ICON
    if not USE_TRAY_ICON:
        return
    if _IconObject:
        try:
            _IconObject.Destroy()
        except:
            pass
        del _IconObject
        _IconObject = None


def init(icons_path, icons_files=None):
    global _IconObject
    global _IconsDict
    global USE_TRAY_ICON
    if not USE_TRAY_ICON:
        return

    if icons_files:
        _IconsDict = icons_files
    else:
        if platform.uname()[0] == 'Linux' or platform.uname()[0] == 'Darwin':
            _IconsDict = _LinuxIcons
        else:
            _IconsDict = _WindowsIcons

    import wx

    from twisted.internet import reactor  # @UnresolvedImport

    def create_menu_item(menu, label, func, icon=None):
        item = wx.MenuItem(menu, -1, label)
        menu.Bind(wx.EVT_MENU, func, id=item.GetId())
        if icon is not None:
            item.SetBitmap(icon)
        menu.AppendItem(item)
        return item

    class MyTaskBarIcon(wx.TaskBarIcon):
        def __init__(self, icons_path, current_icon_name=None):
            super(MyTaskBarIcon, self).__init__()
            self.icons_path = icons_path
            self.icons = {}
            self.popup_icons = {}
            for name, filename in icons_dict().items():
                self.icons[name] = wx.IconFromBitmap(wx.Bitmap(os.path.join(icons_path, filename)))
            for name, filename in popup_icons_dict().items():
                self.popup_icons[name] = wx.Bitmap(os.path.join(icons_path, filename))
            if len(self.icons) == 0:
                self.icons['default'] = ''
            if current_icon_name is not None and current_icon_name in list(self.icons.keys()):
                self.current = current_icon_name
            else:
                self.current = list(self.icons.keys())[0]
            self.Bind(wx.EVT_TASKBAR_LEFT_DOWN, self.on_left_down)
            self.select_icon(self.current)

        def CreatePopupMenu(self):
            menu = wx.Menu()
            create_menu_item(menu, 'open', self.on_show, self.popup_icons.get('open', None))
            create_menu_item(menu, 'synchronize', self.on_sync, self.popup_icons.get('sync', None))
            create_menu_item(menu, 'reconnect', self.on_reconnect, self.popup_icons.get('reconnect', None))
            create_menu_item(menu, 'restart', self.on_restart, self.popup_icons.get('restart', None))
            create_menu_item(menu, 'exit', self.on_exit, self.popup_icons.get('shutdown', None))
            self.menu = menu
            return menu

        def on_left_down(self, event):
            control('show')

        def on_show(self, event):
            control('show')

        def on_sync(self, event):
            control('sync')

        def on_hide(self, event):
            control('hide')

        def on_restart(self, event):
            control('restart')

        def on_exit(self, event):
            control('exit')

        def on_reconnect(self, event):
            control('reconnect')

        def on_toolbar(self, event):
            control('toolbar')

        def select_icon(self, icon_name):
            # print 'select_icon', icon_name, self.icons
            if icon_name in list(self.icons.keys()):
                self.current = icon_name
                self.SetIcon(self.icons.get(self.current, list(self.icons.values())[0]), LABEL)

        def clear_icon(self):
            self.RemoveIcon()

    class MyApp(wx.App):
        def __init__(self, icons_path):
            self.icons_path = icons_path
            wx.App.__init__(self, False)

        def OnInit(self):
            self.trayicon = MyTaskBarIcon(self.icons_path)
            print('OnInit')
            return True

        def OnExit(self):
            print('OnExit')
            try:
                self.trayicon.Destroy()
            except:
                pass

        def SetIcon(self, name):
            # if self.trayicon.IsAvailable():
            print('try')
            self.trayicon.select_icon(name)
            print('ok')

        def Stop(self):
            self.trayicon.clear_icon()
            try:
                self.trayicon.Destroy()
            except:
                pass

    _IconObject = MyApp(icons_path)
    reactor.registerWxApp(_IconObject)
    reactor.addSystemEventTrigger('before', 'shutdown', main_porcess_stopped)


def main_porcess_stopped():
    global _IconObject
    print('main_porcess_stopped', _IconObject)
    if _IconObject:
        try:
            _IconObject.Stop()
        except:
            pass
        del _IconObject
        _IconObject = None


#------------------------------------------------------------------------------


def control(cmd):
    print('control', cmd)
    global _ControlFunc
    if _ControlFunc is not None:
        _ControlFunc(cmd)


def draw_icon(name):
    global _IconObject
    if _IconObject is not None:
        _IconObject.SetIcon(name)


def set_icon(name):
    global _CurrentIcon
    _CurrentIcon = name
    draw_icon(name)


def restore_icon():
    global _CurrentIcon
    draw_icon(_CurrentIcon)


def state_changed(network, p2p):
    # print 'state_changed', network, p2p
    if [
        network,
        p2p,
    ].count('CONNECTED') == 2:
        set_icon('connected')
        return
    # if network == 'DISCONNECTED':
    #     set_icon('off')
    #     return
    # if network == 'CONECTED' and p2p == 'DISCONNECTED':
    #     set_icon('red')
    #     return
    # if [ network, p2p ].count('CONNECTED') == 1:
    #     set_icon('yellow')
    #     return
    set_icon('off')


def SetControlFunc(f):
    global _ControlFunc
    _ControlFunc = f


#------------------------------------------------------------------------------

if __name__ == '__main__':

    def test_control(cmd):
        print('test_control', cmd)
        if cmd == 'exit':
            reactor.stop()
            print('reactor stopped')
            # os._exit(0)
            # sys.exit()

    from twisted.internet import wxreactor
    wxreactor.install()
    from twisted.internet import reactor  # @UnresolvedImport
    init(sys.argv[1])
    SetControlFunc(test_control)
    reactor.run()



# From https://github.com/davidlazar/python-drbg

import hashlib
import hmac


class DRBG(object):
    def __init__(self, seed):
        self.key = b'\x00'*64
        self.val = b'\x01'*64
        self.reseed(seed)

    def hmac(self, key, val):
        return hmac.new(key, val, hashlib.sha512).digest()

    def reseed(self, data=b''):
        self.key = self.hmac(self.key, self.val + b'\x00' + data)
        self.val = self.hmac(self.key, self.val)
        if data:
            self.key = self.hmac(self.key, self.val + b'\x01' + data)
            self.val = self.hmac(self.key, self.val)

    def generate(self, n):
        xs = b''
        while len(xs) < n:
            self.val = self.hmac(self.key, self.val)
            xs += self.val

        self.reseed()
        return xs[:n]



#!/usr/bin/env python
# rebuild.py
#
# Copyright (C) 2008 Stanislav Evseev, Veselin Penev  https://bitdust.io
#
# This file (rebuild.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import
import six

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import time
import traceback
import multiprocessing

from collections import OrderedDict

from threading import Thread, Lock

#------------------------------------------------------------------------------

from bitdust.logs import lg

#------------------------------------------------------------------------------

_WorkerQueue = multiprocessing.Queue()
_WorkerLock = Lock()

#------------------------------------------------------------------------------


class my_decorator_class(object):

    def __init__(self, target):
        self.target = target

    def __call__(self, *args, **kwargs):
        task_id = args[-1]
        args = args[:-1]
        res = None

        try:
            res = self.target(*args)
        except Exception as e:
            traceback.print_exc()
            res = e
        finally:
            try:
                _WorkerQueue.put(1)
            except Exception:
                print('worker got Exception in queue -- exiting')

        return res


def func_thread(tasks, pool):
    while True:
        if _Debug:
            lg.out(_DebugLevel, 'raid.worker.func_thread tasks: %d' % len(tasks))

        try:
            _WorkerQueue.get()
        except (EOFError, OSError, IOError):
            print('publisher got EOFError or OSError or IOError -- exiting')
            break

        with _WorkerLock:
            try:
                task = tasks.popitem(last=False)
            except KeyError:
                task = None

        if task:
            func, params, callback, error_callback, task_id = task[1]
            if _Debug:
                lg.out(_DebugLevel, 'raid.worker.func_thread is going to apply task %s with callback %r' % (
                    task_id,
                    callback,
                ))
            if six.PY3:
                pool.apply_async(func=func, args=params + (task_id, ), callback=callback, error_callback=error_callback)
            else:
                pool.apply_async(func=func, args=params + (task_id, ), callback=callback)
        else:
            try:
                _WorkerQueue.put(1)
            except Exception:
                print('publisher got Exception with queue -- exiting')
                break

            time.sleep(0.1)

    pool.terminate()


class Task(object):

    def __init__(self, task_id):
        self.task_id = task_id
        if _Debug:
            lg.out(_DebugLevel, 'raid.worker.Task created  task_id=%s' % task_id)

    @property
    def tid(self):
        return self.task_id


class Manager(object):

    def __init__(self, ncpus):
        self._ncpus = ncpus

        if six.PY34:
            try:
                multiprocessing.set_start_method('spawn')
            except RuntimeError:
                pass

        multiprocessing.util.log_to_stderr(multiprocessing.util.SUBDEBUG)  # @UndefinedVariable

        from bitdust.system import bpio
        if bpio.Windows():
            from bitdust.system import deploy
            deploy.init_base_dir()
            venv_python_path = os.path.join(deploy.current_base_dir(), 'venv', 'Scripts', 'bitdust-node.exe')
            lg.info('will use %s as multiprocessing executable' % venv_python_path)
            multiprocessing.set_executable(venv_python_path)

        self.processor = multiprocessing.Pool(ncpus)

        self.tasks = OrderedDict({})
        self.task_id = 0

        self.thread = Thread(target=func_thread, args=(self.tasks, self.processor))
        self.thread.daemon = True
        self.thread.start()

        self._propagate_queue()

    def _propagate_queue(self):
        for i in range(self.ncpus):
            _WorkerQueue.put(1)

    def submit(self, func, params, callback, error_callback=None):
        with _WorkerLock:
            self.task_id += 1
            self.tasks[self.task_id] = (my_decorator_class(target=func), params, callback, error_callback, self.task_id)
        return Task(self.task_id)

    @property
    def ncpus(self):
        return self._ncpus

    def terminate(self):
        self.processor.terminate()

    def cancel(self, task_id):
        with _WorkerLock:
            try:
                del self.tasks[task_id]
            except KeyError:
                pass



#!/usr/bin/python
# service_joint_postman.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_joint_postman.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_joint_postman
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return JointPostmanService()


class JointPostmanService(LocalService):

    service_name = 'service_joint_postman'
    config_path = 'services/joint-postman/enabled'
    data_dir_required = True

    last_time_keys_synchronized = None

    def dependent_on(self):
        return [
            'service_private_messages',
            'service_customer_family',
        ]

    def start(self):
        from bitdust.stream import postman
        postman.init()
        return True

    def stop(self):
        from bitdust.stream import postman
        postman.shutdown()
        return True

    def request(self, json_payload, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        from bitdust.stream import postman
        from bitdust.userid import global_id
        try:
            action = json_payload['action']
        except:
            lg.exc()
            return p2p_service.SendFail(newpacket, 'invalid payload')
        # TODO: validate signature and the key
        if action == 'queue-connect':  #  or action == 'queue-connect-follow':
            try:
                consumer_id = global_id.latest_glob_id(json_payload.get('consumer_id'))
                producer_id = global_id.latest_glob_id(json_payload.get('producer_id'))
                group_key_info = json_payload.get('group_key')
            except:
                lg.warn('wrong payload: %r' % json_payload)
                return p2p_service.SendFail(newpacket, 'wrong payload')
            result_defer = Deferred()
            postman.on_queue_connect_request(newpacket, result_defer, consumer_id, producer_id, group_key_info)
            return result_defer
        lg.warn('wrong action request' % newpacket.Payload)
        return p2p_service.SendFail(newpacket, 'wrong action request')

    def cancel(self, json_payload, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        from bitdust.stream import postman
        try:
            action = json_payload['action']
            queue_id = json_payload.get('queue_id', None)
            consumer_id = json_payload['consumer_id']
            producer_id = json_payload['producer_id']
            group_key_info = json_payload['group_key']
        except:
            lg.warn('wrong payload: %r' % json_payload)
            return p2p_service.SendFail(newpacket, 'wrong payload')
        # TODO: validate signature and the key
        if action == 'queue-disconnect':
            result_defer = Deferred()
            postman.on_queue_disconnect_request(newpacket, result_defer, consumer_id, producer_id, group_key_info, queue_id)
            return result_defer
        lg.warn('wrong action request' % newpacket.Payload)
        return p2p_service.SendFail(newpacket, 'wrong action request')



#!/usr/bin/python
# config.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (config.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: config
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys
import re

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.append(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..'))

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

#------------------------------------------------------------------------------

_Config = None

#------------------------------------------------------------------------------


def init(configDir):
    global _Config
    if _Config is None:
        _Config = DetailedConfig(configDir)
        if _Debug:
            lg.out(_DebugLevel, 'config.init     at %s' % configDir)
    else:
        lg.warn('already called, was set up in %s' % _Config.getConfigDir())


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'config.shutdown')
    global _Config
    if _Config:
        del _Config
        _Config = None


#------------------------------------------------------------------------------


def conf():
    global _Config
    return _Config


#------------------------------------------------------------------------------


class BaseConfig(object):
    def __init__(self, configDir):
        self.configDir = configDir

    def getConfigDir(self):
        return self.configDir

    def exist(self, entryPath):
        return self._get(entryPath) is not None

    def registered(self, entryPath):
        return self._exists(entryPath)

    def remove(self, entryPath):
        elemList = self._parseEntryPath(entryPath)
        fpath = os.path.join(self.configDir, *elemList)
        return self._rmrf(fpath)

    def listEntries(self, entryPath):
        entries = self._get(entryPath)
        t = type(entries)
        if t is bytes:
            # lg.warn( 'argument to listEntries is a file: %s' % entryPath )
            return []
        if entries is None:
            return []
        assert t is list
        return entries

    def hasChilds(self, entryPath):
        elemList = self._parseEntryPath(entryPath)
        fpath = os.path.join(self.configDir, *elemList)
        if os.path.isdir(fpath):
            return True
        return False

    def listAllEntries(self):
        try:
            from bitdust.system import bpio
            l = []
            abspth = bpio.portablePath(os.path.abspath(self.getConfigDir()))
            for subpath in bpio.list_dir_recursive(abspth):
                path = bpio.portablePath(subpath)
                l.append(path.replace(abspth, '').strip('/'))
            return l
        except:
            lg.exc()
            return []

    def getData(self, entryPath, default=None):
        data = self._get(entryPath)
        if data is None:
            return default
        if isinstance(data, list):
            # lg.warn('argument to getData is a directory: %s' % entryPath)
            return default
        return data

    def setData(self, entryPath, value):
        return self._set(entryPath, value)

    def getInt(self, entryPath, default=None):
        s = self.getData(entryPath)
        if s is None:
            return default
        try:
            s = str(s).strip().strip('"')
            return int(s)
        except ValueError:
            return default

    def setInt(self, entryPath, value):
        return self._set(entryPath, str(int(value)))

    def getFloat(self, entryPath, default=None):
        s = self.getData(entryPath)
        if s is None:
            return default
        try:
            s = str(s).strip().strip('"')
            return float(s)
        except ValueError:
            return default

    def setFloat(self, entryPath, value):
        return self._set(entryPath, str(float(value)))

    def getBool(self, entryPath, default=None):
        data = self.getData(entryPath)
        if data is None:
            return default
        return True if str(data).strip() in [
            'True',
            'true',
            '1',
            'on',
        ] else False

    def setBool(self, entryPath, value):
        return self._set(entryPath, 'true' if bool(value) else 'false')

    def getString(self, entryPath, default=''):
        data = self.getData(entryPath)
        if data is None:
            return default
        data = str(data).strip()
        if len(data) < 2:
            return default
        if data[0] == data[-1] == '"':
            data = data[1:-1]
        try:
            out = []
            i = 0
            while i < len(data):
                c = data[i]
                if c == '\\':
                    out.append(data[i + 1])
                    i += 2
                else:
                    out.append(c)
                    i += 1
            return ''.join(out)
        except IndexError:
            return default

    def setString(self, entryPath, value):
        out = ['"']
        for x in str(value):
            if x in '\\"':
                out.append('\\')
            out.append(x)
        out.append('"')
        return self._set(entryPath, ''.join(out))

    def _mkdir(self, dpath):
        if os.path.isfile(dpath):
            self._unlink(dpath)
        if not os.path.isdir(dpath):
            try:
                if sys.platform == 'win32':
                    os.makedirs(dpath)
                else:
                    os.makedirs(dpath, 0o700)
                return True
            except OSError:
                lg.exc('error creating directory: %s' % dpath)
                return False
        return True

    def _listdir(self, dpath):
        try:
            return os.listdir(dpath)
        except OSError:
            lg.exc('error listing dir: %s' % dpath)
            return []

    def _unlink(self, fpath):
        try:
            if os.path.isdir(fpath):
                os.rmdir(fpath)
            else:
                os.unlink(fpath)
            return True
        except OSError:
            lg.exc('error unlinking: %s' % fpath)
            return False

    def _rmrf(self, fpath):
        if os.path.isdir(fpath):
            files = self._listdir(fpath)
            for f in files:
                childPath = os.path.join(fpath, f)
                self._rmrf(childPath)
            return self._unlink(fpath)
        elif os.path.isfile(fpath):
            return self._unlink(fpath)
        return False

    def _validateElemList(self, elemList):
        for x in elemList:
            assert x.strip() == x
            assert '\\' not in x
            assert '/' not in x
            assert x not in ('.', '..')

    def _parseEntryPath(self, entryPath):
        elemList = entryPath.split('/')
        if elemList and (not elemList[0]):
            del elemList[0]
        if elemList:
            assert elemList[-1]
        self._validateElemList(elemList)
        return elemList

    def _exists(self, entryPath):
        elemList = self._parseEntryPath(entryPath)
        fpath = os.path.join(self.configDir, *elemList)
        if os.path.isdir(fpath):
            return True
        elif os.path.isfile(fpath):
            return True
        return False

    def _get(self, entryPath):
        elemList = self._parseEntryPath(entryPath)
        fpath = os.path.join(self.configDir, *elemList)
        if os.path.isdir(fpath):
            out = []
            for x in self._listdir(fpath):
                childPath = '/'.join(elemList + [x])
                out.append(childPath)
            return out
        elif os.path.isfile(fpath):
            data = None
            try:
                f = open(fpath, 'rb')
                data = strng.to_text(f.read())
                f.close()
            except (OSError, IOError):
                lg.exc('error reading from file: %s' % fpath)
            return data
        return None

    def _set(self, entryPath, data):
        elemList = self._parseEntryPath(entryPath)
        assert elemList
        dpath = self.configDir
        for d in elemList[:-1]:
            dpath = os.path.join(dpath, d)
            self._mkdir(dpath)
        fpath = os.path.join(dpath, elemList[-1])
        s = strng.to_bin(data)
        try:
            f = open(fpath, 'wb')
            f.write(s)
            f.close()
            return True
        except (OSError, IOError):
            lg.exc('error writing to file: %s' % fpath)
        return False


#------------------------------------------------------------------------------


class DefaultsConfig(BaseConfig):
    _default = {}

    def setDefaultValue(self, entryPath, value):
        self._default[entryPath] = value

    def getDefaultValue(self, entryPath):
        return self._default.get(entryPath, None)

    def getData(self, entryPath, default=None):
        if default is not None:
            return BaseConfig.getData(self, entryPath, default)
        result = BaseConfig.getData(self, entryPath)
        if result is not None:
            return result
        return self.getDefaultValue(entryPath)

    def getOriginalData(self, entryPath):
        return BaseConfig.getData(self, entryPath)


#------------------------------------------------------------------------------


class NotifiableConfig(DefaultsConfig):
    def __init__(self, configDir):
        super(NotifiableConfig, self).__init__(configDir)
        self.callbacks = {}

    def addConfigNotifier(self, mask, cb):
        """
        You can add a callback to catch a moment when some particular option
        were modified. Mask is a string which is used to compared in this way:

            entryPath.startswith(mask)

        The callback will be fired with such arguments:

            cb(entryPath, newValue, oldValue, result)
        """
        if mask not in self.callbacks:
            self.callbacks[mask] = []
        self.callbacks[mask].append(cb)

    def removeConfigNotifier(self, mask, cb=None):
        """
        Remove existing callback.
        """
        if cb and mask in self.callbacks:
            self.callbacks[mask].remove(cb)
        else:
            self.callbacks.pop(mask, None)

    def _set(self, entryPath, newValue):
        oldValue = self._get(entryPath)
        result = BaseConfig._set(self, entryPath, newValue)
        for mask, cb_list in self.callbacks.items():
            if entryPath.startswith(mask):
                for cb in cb_list:
                    cb(entryPath, newValue, oldValue, result)
        return result


#------------------------------------------------------------------------------


class FixedTypesConfig(NotifiableConfig):
    # TODO: verify value against type of the field when populating every option

    def __init__(self, configDir):
        super(FixedTypesConfig, self).__init__(configDir)
        try:
            from bitdust.main import config_types
            self._types = config_types.defaults()
            self._labels = config_types.labels()
        except:
            self._types = {}
            self._labels = {}

    def listKnownTypes(self):
        return list(self._types.keys())

    def setType(self, key, typ):
        self._types[key] = typ

    def getType(self, entryPath):
        return self._types.get(entryPath, 0)

    def getTypeLabel(self, entryPath):
        return self._labels.get(self.getType(entryPath))

    def getTypeMetaInfo(self, entryPath):
        from bitdust.main import config_types
        typ = self.getType(entryPath)
        if not typ:
            return {}
        if typ in [
            config_types.TYPE_STRING,
            config_types.TYPE_TEXT,
            config_types.TYPE_UNDEFINED,
            config_types.TYPE_PASSWORD,
            config_types.TYPE_INTEGER,
            config_types.TYPE_BOOLEAN,
            config_types.TYPE_FLOATING_POINT,
            config_types.TYPE_POSITIVE_FLOATING_POINT,
            config_types.TYPE_NON_ZERO_POSITIVE_FLOATING_POINT,
        ]:
            return {}
        if typ == config_types.TYPE_POSITIVE_INTEGER:
            return {
                'min': 0,
            }
        if typ == config_types.TYPE_PORT_NUMBER:
            return {
                'min': 1,
                'max': 65535,
            }
        if typ == config_types.TYPE_NON_ZERO_POSITIVE_INTEGER:
            return {
                'min': 1,
            }
        if typ in [
            config_types.TYPE_FOLDER_PATH,
            config_types.TYPE_FILE_PATH,
        ]:
            # TODO: to be decided later
            return {}
        elif typ == config_types.TYPE_COMBO_BOX:
            if entryPath == 'services/customer/suppliers-number':
                from bitdust.raid import eccmap
                return {
                    'possible_values': eccmap.SuppliersNumbers(),
                }
            else:
                raise TypeError('unexpected option type for %r' % entryPath)
        return {}

    def getValueOfType(self, entryPath, default=None):
        from bitdust.main import config_types
        typ = self.getType(entryPath)
        value = None
        if not typ or typ in [
            config_types.TYPE_STRING,
            config_types.TYPE_TEXT,
            config_types.TYPE_UNDEFINED,
        ]:
            value = self.getString(entryPath, default=default)
        elif typ in [
            config_types.TYPE_BOOLEAN,
        ]:
            value = self.getBool(entryPath, default=default)
        elif typ in [
            config_types.TYPE_INTEGER,
            config_types.TYPE_POSITIVE_INTEGER,
            config_types.TYPE_NON_ZERO_POSITIVE_INTEGER,
            config_types.TYPE_PORT_NUMBER,
        ]:
            value = self.getInt(entryPath, default=default)
        elif typ in [
            config_types.TYPE_FOLDER_PATH,
            config_types.TYPE_FILE_PATH,
            config_types.TYPE_PASSWORD,
        ]:
            value = self.getString(entryPath, default=default) or ''
            if typ in [
                config_types.TYPE_FOLDER_PATH,
            ]:
                if value:
                    from bitdust.system import bpio
                    if bpio.Windows():
                        if not value.endswith(u'/') and not value.endswith(u'\\'):
                            value = value + u'\\'
                    elif not value.endswith(u'/'):
                        value = value + u'/'
        elif typ == config_types.TYPE_COMBO_BOX:
            if entryPath == 'services/customer/suppliers-number':
                value = self.getInt(entryPath, default=default)
            else:
                raise TypeError('unexpected option type for %r' % entryPath)
        elif typ in [
            config_types.TYPE_FLOATING_POINT,
            config_types.TYPE_POSITIVE_FLOATING_POINT,
            config_types.TYPE_NON_ZERO_POSITIVE_FLOATING_POINT,
        ]:
            value = self.getFloat(entryPath, default=default)
        else:
            value = self.getData(entryPath, default=default)
        return value

    def setValueOfType(self, entryPath, value):
        from bitdust.main import config_types
        typ = self.getType(entryPath)
        if not typ or typ in [
            config_types.TYPE_STRING,
            config_types.TYPE_TEXT,
            config_types.TYPE_UNDEFINED,
        ]:
            self.setString(entryPath, strng.text_type(value))
        elif typ in [
            config_types.TYPE_BOOLEAN,
        ]:
            if strng.is_string(value):
                vl = strng.to_text(value).strip().lower() in ['true', '1', 'on']
            else:
                vl = bool(value)
            self.setBool(entryPath, vl)
        elif typ in [
            config_types.TYPE_INTEGER,
            config_types.TYPE_POSITIVE_INTEGER,
            config_types.TYPE_NON_ZERO_POSITIVE_INTEGER,
            config_types.TYPE_PORT_NUMBER,
        ]:
            self.setInt(entryPath, value)
        elif typ in [
            config_types.TYPE_FOLDER_PATH,
            config_types.TYPE_FILE_PATH,
            config_types.TYPE_PASSWORD,
        ]:
            self.setString(entryPath, value)
        elif typ == config_types.TYPE_COMBO_BOX:
            if entryPath == 'services/customer/suppliers-number':
                value = self.setInt(entryPath, int(value))
            else:
                raise TypeError('unexpected option type for %r' % entryPath)
        elif typ in [
            config_types.TYPE_FLOATING_POINT,
            config_types.TYPE_POSITIVE_FLOATING_POINT,
            config_types.TYPE_NON_ZERO_POSITIVE_FLOATING_POINT,
        ]:
            self.setFloat(entryPath, value)
        else:
            self.setData(entryPath, strng.text_type(value))
        return True


#------------------------------------------------------------------------------


class CachedConfig(FixedTypesConfig):

    _cache = {}

    def _set(self, entryPath, data):
        if entryPath in self._cache:
            if self._cache[entryPath] == data:
                return True
        self._cache[entryPath] = data
        result = FixedTypesConfig._set(self, entryPath, data)
        return result

    def _get(self, entryPath):
        if entryPath in self._cache:
            return self._cache[entryPath]
        result = FixedTypesConfig._get(self, entryPath)
        self._cache[entryPath] = result
        return result

    def cache(self):
        return self._cache

    def reloadCache(self):
        """
        Reload whole cache from local files.
        """
        # TODO

    def storeCache(self):
        """
        Write all cached entries into local files.
        """
        # TODO


#------------------------------------------------------------------------------


class DetailedConfig(CachedConfig):

    _labels = {}
    _read_only = {}
    _infos = {}

    def __init__(self, configDir):
        super(DetailedConfig, self).__init__(configDir)
        try:
            from bitdust.main import config_details
            self._load_details(config_details.raw())
        except:
            lg.exc()

    def _load_details(self, src):
        current_option = ''
        for line in src.splitlines():
            if not line.strip():
                continue
            ro = re.match('^\[(.+?)\](.*?)$', line)
            if ro:
                self._read_only[ro.group(1).strip()] = True
                line = line.replace('[', '{').replace(']', '}')
            r = re.match('^\{(.+?)\}(.*?)$', line)
            if r:
                current_option = r.group(1).strip()
                self._labels[current_option] = r.group(2).strip()
            else:
                if current_option:
                    if current_option not in self._infos:
                        self._infos[current_option] = ''
                    self._infos[current_option] += line.strip() + '\n'

    def getLabel(self, entryPath):
        return self._labels.get(entryPath, '') or entryPath.split('/')[-1]

    def getInfo(self, entryPath):
        return (self._infos.get(entryPath, '') or '').strip(' \n\t')

    def getReadOnly(self, entryPath):
        return self._read_only.get(entryPath, False) or False

    def toJson(self, entryPath, include_info=True):
        result = {
            'key': entryPath,
            'value': self.getValueOfType(entryPath),
            'type': self.getTypeLabel(entryPath),
            'label': self.getLabel(entryPath),
            'readonly': self.getReadOnly(entryPath),
            'default': self.getDefaultValue(entryPath),
        }
        if include_info:
            result['info'] = self.getInfo(entryPath)
        result.update(self.getTypeMetaInfo(entryPath))
        return result


#------------------------------------------------------------------------------


def main():
    """
    Read settings from 'config' file and prints values from your queries to stdout.
    """
    from bitdust.logs import lg
    lg.set_debug_level(24)
    from bitdust.main import settings
    settings.init()
    init(settings.ConfigDir())
    print(conf().listEntries(''))
    try:
        inp = sys.argv[1].rstrip('/')
    except:
        print('wrong input')
        return
    if not conf().exist(inp):
        print('not exist')
        return
    if not conf().hasChilds(inp):
        print(inp, conf().getData(inp))
        return
    for child in conf().listEntries(inp):
        if conf().hasChilds(child):
            print(child, conf().listEntries(child))
        else:
            print(child, conf().getData(child))
    settings.shutdown()
    return


#    last = ''
#    for entry in sorted(conf()._types.keys()):
#        parent, key = entry.rsplit('/', 1)
#        while parent.count('/'):
#            p, parent = parent.rsplit('/', 1)
#            if p != 'services':
#                print p
#            parent = '  ' + parent
#        if parent != last:
#            print parent
#            last = parent
#        print ' ' * (last.count(' ') + 1) * 2, key, '\t\t\t\t', conf().get_type_label(entry).upper()
# print '\n'.join(map(lambda x: "    '%s':\t\t\tNode," % x, sorted(conf().listAllEntries())))
# s = conf().getData('details')
# conf()._load_details(s)

if __name__ == '__main__':
    main()



#!/usr/bin/python
# backup_fs.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_fs.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup_fs.

This is some kind of file system.
To store backed up data on remote peers we can not use original files and folders names -
they must be encrypted or indexed. I decide to use index, but keep the files and folders structure.
Instead of names I am using numbers.

For example::

  C:/Documents and Settings/veselin/Application Data/Google/

Can have a path ID like this::

  0/2/0/5/23

Linux paths can be indexed same way::

  /home/veselin/Documents/document.pdf

Can be translated to::

  0/2/4/18

The software keeps 2 index dictionaries in the memory:

* path -> ID
* ID -> path

Those dictionaries are trees - replicates the file system structure.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range
from io import StringIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys
import time
import json
import random

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.lib import strng

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings
from bitdust.main import listeners

from bitdust.services import driver

from bitdust.lib import misc
from bitdust.lib import packetid
from bitdust.lib import jsn

from bitdust.crypt import my_keys

from bitdust.contacts import identitycache

from bitdust.interface import api

from bitdust.userid import my_id
from bitdust.userid import global_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

INFO_KEY = 'i'
UNKNOWN = -1
FILE = 0
DIR = 1
TYPES = {
    UNKNOWN: 'UNKNOWN',
    FILE: 'FILE',
    DIR: 'DIR',
}

#------------------------------------------------------------------------------

_FileSystemIndexByName = {}
_FileSystemIndexByID = {}
_RevisionNumber = {}
_ItemsCount = 0
_FilesCount = 0
_DirsCount = 0
_SizeFiles = 0
_SizeFolders = 0
_SizeBackups = 0

#------------------------------------------------------------------------------


def init():
    """
    Some initial steps can be done here.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_fs.init')
    LoadAllIndexes()
    SaveIndex()


def shutdown():
    """
    Should be called when the program is finishing.
    """
    if _Debug:
        lg.out(_DebugLevel, 'backup_fs.shutdown')
    ClearAllIndexes()


#------------------------------------------------------------------------------


def fs(customer_idurl=None, key_alias='master'):
    """
    Access method for forward index: [path] -> [ID].
    """
    global _FileSystemIndexByName
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _FileSystemIndexByName:
        _FileSystemIndexByName[customer_idurl] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new customer registered : %r' % customer_idurl)
    if key_alias is None:
        return _FileSystemIndexByName[customer_idurl]
    if key_alias not in _FileSystemIndexByName[customer_idurl]:
        _FileSystemIndexByName[customer_idurl][key_alias] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new key alias registered for customer %r : %r' % (customer_idurl, key_alias))
    return _FileSystemIndexByName[customer_idurl][key_alias]


def fsID(customer_idurl=None, key_alias='master'):
    """
    Access method for backward index: [ID] -> [path].
    """
    global _FileSystemIndexByID
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _FileSystemIndexByID:
        _FileSystemIndexByID[customer_idurl] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new customer registered : %r' % customer_idurl)
    if key_alias is None:
        return _FileSystemIndexByID[customer_idurl]
    if key_alias not in _FileSystemIndexByID[customer_idurl]:
        _FileSystemIndexByID[customer_idurl][key_alias] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new key alias registered for customer %r : %r' % (customer_idurl, key_alias))
    return _FileSystemIndexByID[customer_idurl][key_alias]


#------------------------------------------------------------------------------


def revision(customer_idurl=None, key_alias='master'):
    """
    Mutator method to access current software revision number.
    """
    global _RevisionNumber
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _RevisionNumber:
        _RevisionNumber[customer_idurl] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new customer registered : %r' % customer_idurl)
    if key_alias not in _RevisionNumber[customer_idurl]:
        _RevisionNumber[customer_idurl][key_alias] = -1
        if _Debug:
            lg.dbg(_DebugLevel, 'new key alias registered for customer %r : %r' % (customer_idurl, key_alias))
    return _RevisionNumber[customer_idurl][key_alias]


def commit(new_revision_number=None, customer_idurl=None, key_alias='master'):
    """
    Need to be called after any changes in the index database.

    This increase revision number by 1 or set revision to ``new_revision_number`` if not None.
    """
    global _RevisionNumber
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _RevisionNumber:
        _RevisionNumber[customer_idurl] = {}
        if _Debug:
            lg.dbg(_DebugLevel, 'new customer registered : %r' % customer_idurl)
    if key_alias not in _RevisionNumber[customer_idurl]:
        _RevisionNumber[customer_idurl][key_alias] = 0
        if _Debug:
            lg.dbg(_DebugLevel, 'new key alias registered for customer %r : %r' % (customer_idurl, key_alias))
    old_v = _RevisionNumber[customer_idurl][key_alias]
    if new_revision_number is not None:
        _RevisionNumber[customer_idurl][key_alias] = new_revision_number
    else:
        _RevisionNumber[customer_idurl][key_alias] += 1
    new_v = _RevisionNumber[customer_idurl][key_alias]
    if _Debug:
        lg.args(_DebugLevel, old=old_v, new=new_v, c=customer_idurl, k=key_alias)
    return old_v, new_v


def forget(customer_idurl=None, key_alias=None):
    """
    Release currently known revision number for the corresponding index.
    """
    global _RevisionNumber
    if _Debug:
        lg.args(_DebugLevel, c=customer_idurl, k=key_alias)
    if customer_idurl is None:
        _RevisionNumber.clear()
        return
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _RevisionNumber:
        lg.warn('customer %r was not registered' % customer_idurl)
        return
    if key_alias is None:
        _RevisionNumber[customer_idurl].clear()
        return
    if key_alias not in _RevisionNumber[customer_idurl]:
        lg.warn('key alias %r was not registered for customer %r' % (key_alias, customer_idurl))
        return
    _RevisionNumber[customer_idurl].pop(key_alias)


#------------------------------------------------------------------------------


def known_customers():
    global _FileSystemIndexByID
    return list(_FileSystemIndexByID.keys())


def known_keys_aliases(customer_idurl):
    global _FileSystemIndexByID
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    return list(_FileSystemIndexByID.get(customer_idurl, {}).keys())


#------------------------------------------------------------------------------


def counter():
    """
    Software keeps track of total number of indexed items, this returns that
    value.
    """
    global _ItemsCount
    return _ItemsCount


def numberfiles():
    """
    Number of indexed files.
    """
    global _FilesCount
    return _FilesCount


def numberfolders():
    """
    Number of indexed files.
    """
    global _DirsCount
    return _DirsCount


def sizefiles():
    """
    Total size of all indexed files.
    """
    global _SizeFiles
    return _SizeFiles


def sizefolders():
    """
    Total size of all indexed folders.

    May be incorrect, because folder size is not calculated regular yet.
    """
    global _SizeFolders
    return _SizeFolders


def sizebackups():
    """
    Total size of all indexed backups.
    """
    global _SizeBackups
    return _SizeBackups


#------------------------------------------------------------------------------


class FSItemInfo():

    """
    A class to represent a remote file or folder.
    """

    def __init__(self, name='', path_id='', typ=UNKNOWN, key_id=None):
        self.unicodename = strng.to_text(name)
        self.path_id = path_id
        self.type = typ
        self.size = -1
        self.key_id = key_id
        self.versions = {}

    def __repr__(self):
        return '<%s %s %d %s>' % (TYPES[self.type], misc.unicode_to_str_safe(self.name()), self.size, self.key_id)

    def to_json(self):
        return {
            'name': self.unicodename,
            'path_id': self.path_id,
            'type': self.type,
            'size': self.size,
            'key_id': self.key_id,
            'versions': self.versions,
        }

    def filename(self):
        return os.path.basename(self.unicodename)

    def name(self):
        return self.unicodename

    def key_alias(self):
        if not self.key_id:
            return 'master'
        return self.key_id.split('$')[0]

    def exist(self):
        return self.size != -1

    def set_size(self, sz):
        self.size = sz

    def read_stats(self, path):
        if not bpio.pathExist(path):
            return False
        if bpio.pathIsDir(path):
            return False
        try:
            s = os.stat(path)
        except:
            try:
                s = os.stat(path.decode('utf-8'))
            except:
                lg.exc()
                return False
        self.size = int(s.st_size)
        return True

    def read_versions(self, local_path):
        path = bpio.portablePath(local_path)
        if not bpio.pathExist(path):
            return 0
        if not os.access(path, os.R_OK):
            return 0
        totalSize = 0
        for version in bpio.list_dir_safe(path):
            if self.get_version_info(version)[0] >= 0:
                continue
            versionSize = 0
            maxBlock = -1
            if not packetid.IsCanonicalVersion(version):
                continue
            versionpath = os.path.join(path, version)
            if not bpio.pathExist(versionpath):
                continue
            if not os.access(versionpath, os.R_OK):
                return 0
            for filename in bpio.list_dir_safe(versionpath):
                filepath = os.path.join(versionpath, filename)
                if not packetid.IsPacketNameCorrect(filename):
                    lg.warn('incorrect file name found: %s' % filepath)
                    continue
                try:
                    blockNum, supplierNum, _ = filename.split('-')
                    blockNum, supplierNum = int(blockNum), int(supplierNum)
                except:
                    lg.warn('incorrect file name found: %s' % filepath)
                    continue
                try:
                    sz = int(os.path.getsize(filepath))
                except:
                    lg.exc()
                    sz = 0
                # TODO:
                # add some bytes because on remote machines all files are stored as signed.Packet()
                # so they have a header and files size will be bigger than on local machine
                # we do not know how big head is, so just add an approximate value
                versionSize += sz + 1024
                maxBlock = max(maxBlock, blockNum)
            self.set_version_info(version, maxBlock, versionSize)
            totalSize += versionSize
        return totalSize

    def add_version(self, version):
        self.versions[version] = [-1, -1]

    def set_version_info(self, version, maxblocknum, sizebytes):
        self.versions[version] = [maxblocknum, sizebytes]

    def get_version_info(self, version):
        return self.versions.get(version, [-1, -1])

    def get_version_size(self, version):
        return self.versions.get(version, [-1, -1])[1]

    def delete_version(self, version):
        self.versions.pop(version, None)

    def has_version(self, version):
        return version in self.versions

    def any_version(self):
        return len(self.versions) > 0

    def list_versions(self, sorted=False, reverse=False):
        if sorted:
            return misc.sorted_versions(list(self.versions.keys()), reverse)
        return list(self.versions.keys())

    def get_versions(self):
        return self.versions

    def get_latest_version(self):
        if len(self.versions) == 0:
            return None
        return self.list_versions(True)[0]

    def pack_versions(self):
        out = []
        for version in self.list_versions(True):
            info = self.versions[version]
            out.append(version + ':' + str(info[0]) + ':' + str(info[1]))
        return ' '.join(out)

    def unpack_versions(self, inpt):
        for word in inpt.split(' '):
            if not word.strip():
                continue
            try:
                version, maxblock, sz = word.split(':')
                maxblock, sz = int(maxblock), int(sz)
            except:
                version, maxblock, sz = word, -1, -1
            self.set_version_info(version, maxblock, sz)

    def serialize(self, encoding='utf-8', to_json=False):
        if _Debug:
            lg.args(_DebugLevel, k=self.key_id, pid=self.path_id, v=list(self.versions.keys()), i=id(self), iv=id(self.versions))
        if to_json:
            return {
                'n': strng.to_text(self.unicodename, encoding=encoding),
                'i': strng.to_text(self.path_id),
                't': self.type,
                's': self.size,
                'k': self.key_id,
                'v': [{
                    'n': v,
                    'b': self.versions[v][0],
                    's': self.versions[v][1],
                } for v in self.list_versions(sorted=True)],
            }
        e = strng.to_text(self.unicodename, encoding=encoding)
        return '%s %d %d %s\n%s\n' % (self.path_id, self.type, self.size, self.pack_versions(), e)

    def unserialize(self, src, decoding='utf-8', from_json=False):
        if from_json:
            try:
                self.unicodename = strng.to_text(src['n'], encoding=decoding)
                self.path_id = strng.to_text(src['i'], encoding=decoding)
                self.type = src['t']
                self.size = src['s']
                self.key_id = my_keys.latest_key_id(strng.to_text(src['k'], encoding=decoding))
                self.versions = {strng.to_text(v['n']): [v['b'], v['s']] for v in src['v']}
            except:
                lg.exc()
                raise KeyError('Incorrect item format:\n%s' % src)
            if _Debug:
                lg.args(_DebugLevel, k=self.key_id, pid=self.path_id, v=list(self.versions.keys()), i=id(self), iv=id(self.versions))
            return True

        try:
            details, name = strng.to_text(src, encoding=decoding).split('\n')[:2]
        except:
            raise Exception('incorrect item format:\n%s' % src)
        if not details or not name:
            raise Exception('incorrect item format:\n%s' % src)
        try:
            self.unicodename = name
            details = details.split(' ')
            self.path_id, self.type, self.size = details[:3]
            self.type, self.size = int(self.type), int(self.size)
            self.unpack_versions(' '.join(details[3:]))
        except:
            lg.exc()
            raise KeyError('incorrect item format:\n%s' % src)
        if _Debug:
            lg.args(_DebugLevel, k=self.key_id, pid=self.path_id, v=list(self.versions.keys()), i=id(self), iv=id(self.versions))
        return True


#------------------------------------------------------------------------------


def MakeID(itr, randomized=True):
    """
    Create a new unique number for the file or folder to create a index ID.

    Parameter ``itr`` is a reference for a single item in the ``fs()``.
    """
    current_ids = []
    for k in itr.keys():
        if k == 0:
            continue
        if k == settings.BackupIndexFileName():
            continue
        try:
            if isinstance(itr[k], int):
                current_ids.append(int(itr[k]))
            elif isinstance(itr[k], dict) and 0 in itr[k]:
                current_ids.append(int(itr[k][0]))
            else:
                continue
        except:
            lg.exc()
            continue
    new_id = 0
    if randomized:
        digits = 1
        while True:
            attempts = 0
            new_id = int(random.choice('0123456789'))
            while new_id in current_ids and attempts <= 2:
                new_id = int(''.join([v() for v in [
                    lambda: random.choice('0123456789'),
                ]*digits]))
                attempts += 1
            if new_id not in current_ids:
                return new_id
            digits += 1
    while new_id in current_ids:
        new_id += 1
    return new_id


#------------------------------------------------------------------------------


def AddFile(path, read_stats=False, iter=None, iterID=None, key_id=None):
    """
    Scan all components of the ``path`` and create an item in the index for
    that file.

        >>> import backup_fs
        >>> backup_fs.AddFile('C:/Documents and Settings/veselin/Application Data/Google/GoogleEarth/myplaces.kml')
        ('0/0/0/0/0/0/0', {0: 0, u'myplaces.kml': 0}, {'i': <PARENT GoogleEarth -1>, 0: <FILE myplaces.kml -1>})

    Here path must be in "portable" form - only '/' allowed, assume path is a file, not a folder.
    """
    parts = bpio.remotePath(path).split('/')
    key_alias = 'master'
    if key_id:
        key_alias = key_id.split('$')[0]
    if iter is None:
        iter = fs(key_alias=key_alias)
    if iterID is None:
        iterID = fsID(key_alias=key_alias)
    resultID = ''
    parentKeyID = None
    # build whole tree, skip the last part
    for i in range(len(parts) - 1):
        name = parts[i]
        if not name:
            continue
        p = '/'.join(parts[:i + 1])
        if bpio.Linux() or bpio.Mac():
            p = '/' + p
        if name not in iter:
            # made a new ID for this folder, ID starts from 0. new folders will get the last ID +1
            # or it may find a free place in the middle, if some folders or files were removed before
            # this way we try to protect the files and directories names. we store index in the encrypted files
            id = MakeID(iter)
            # build a unique backup id for that file including all indexed ids
            resultID += '/' + str(id)
            # make new sub folder
            ii = FSItemInfo(name=name, path_id=resultID.lstrip('/'), typ=DIR, key_id=(key_id or parentKeyID))
            if read_stats:
                ii.read_stats(p)
            # we use 0 key as decimal value, all files and folders are strings - no conflicts possible 0 != '0'
            iter[ii.name()] = {0: id}
            # also save index from opposite side
            iterID[id] = {INFO_KEY: ii}
        else:
            # get an existing ID from the index
            id = iter[name][0]
            # go down into the existing forest
            resultID += '/' + str(id)
        # move down to the next level
        parentKeyID = iterID[id][INFO_KEY].key_id
        iter = iter[name]
        iterID = iterID[id]
    # the last part of the path is a filename
    filename = parts[-1]
    # make an ID for the filename
    id = MakeID(iter)
    resultID += '/' + str(id)
    resultID = resultID.lstrip('/')
    ii = FSItemInfo(name=filename, path_id=resultID, typ=FILE, key_id=(key_id or parentKeyID))
    if read_stats:
        ii.read_stats(path)
    iter[ii.name()] = id
    iterID[id] = ii
    # finally make a complete backup id - this a relative path to the backed up file
    return resultID, ii, iter, iterID


def AddDir(path, read_stats=False, iter=None, iterID=None, key_id=None, force_path_id=None):
    """
    Add specific local directory to the index, but do not read content of the folder.
    """
    parts = bpio.remotePath(path).split('/')
    force_path_id_parts = []
    if force_path_id is not None:
        force_path_id_parts = bpio.remotePath(force_path_id).split('/')
    key_alias = 'master' if not key_id else key_id.split('$')[0]
    if iter is None:
        iter = fs(key_alias=key_alias)
    if iterID is None:
        iterID = fsID(key_alias=key_alias)
    resultID = ''
    parentKeyID = None
    ii = None
    for i in range(len(parts)):
        name = parts[i]
        if not name:
            continue
        p = '/'.join(parts[:i + 1])
        if bpio.Linux() or bpio.Mac():
            p = '/' + p
        if name not in iter:
            id = 0
            if force_path_id_parts:
                id = int(force_path_id_parts[0])
                force_path_id_parts = force_path_id_parts[1:]
            else:
                id = MakeID(iter)
            resultID += '/' + str(id)
            ii = FSItemInfo(name, path_id=resultID.lstrip('/'), typ=DIR, key_id=(key_id or parentKeyID))
            if read_stats:
                ii.read_stats(p)
            iter[ii.name()] = {0: id}
            iterID[id] = {INFO_KEY: ii}
        else:
            id = iter[name][0]
            resultID += '/' + str(id)
        parentKeyID = iterID[id][INFO_KEY].key_id
        iter = iter[name]
        iterID = iterID[id]
        if i == len(parts) - 1:
            if iterID[INFO_KEY].type != DIR:
                lg.warn('not a dir: %s' % iterID[INFO_KEY])
            iterID[INFO_KEY].type = DIR
    return resultID.lstrip('/'), ii, iter, iterID


def AddLocalPath(localpath, read_stats=False, iter=None, iterID=None, key_id=None):
    """
    Operates like ``AddDir()`` but also recursively reads the entire folder and
    put all items in the index. Parameter ``localpath`` can be a file or folder path.
    """

    def recursive_read_dir(local_path, path_id, iter, iterID):
        c = 0
        lastID = -1
        path = bpio.portablePath(local_path)
        if not os.access(path, os.R_OK):
            return c
        for localname in bpio.list_dir_safe(path):
            p = os.path.join(path, localname)
            name = strng.to_text(localname)
            if bpio.pathIsDir(p):
                if name not in iter:
                    id = MakeID(iter, lastID)
                    ii = FSItemInfo(name=name, path_id=(path_id + '/' + str(id)).lstrip('/'), typ=DIR, key_id=key_id)
                    iter[ii.name()] = {0: id}
                    if read_stats:
                        ii.read_stats(p)
                    iterID[id] = {INFO_KEY: ii}
                    lastID = id
                else:
                    id = iter[name][0]
                c += recursive_read_dir(p, path_id + '/' + str(id), iter[name], iterID[id])
            else:
                id = MakeID(iter, lastID)
                ii = FSItemInfo(name=name, path_id=(path_id + '/' + str(id)).lstrip('/'), typ=FILE, key_id=key_id)
                if read_stats:
                    ii.read_stats(p)
                iter[ii.name()] = id
                iterID[id] = ii
                c += 1
                lastID = id
        return c

    localpath = bpio.portablePath(localpath)
    if bpio.pathIsDir(localpath):
        path_id, itemInfo, iter, iterID = AddDir(localpath, read_stats=read_stats, iter=iter, iterID=iterID, key_id=key_id)
        num = recursive_read_dir(localpath, path_id, iter, iterID)
        return path_id, iter, iterID, num
    else:
        path_id, itemInfo, iter, iterID = AddFile(localpath, read_stats=read_stats, iter=iter, iterID=iterID, keyID=key_id)
        return path_id, iter, iterID, 1
    return None, None, None, 0


def PutItem(name, parent_path_id, as_folder=False, iter=None, iterID=None, key_id=None):
    """
    Acts like AddFile() but do not follow the directory structure. This just
    "bind" some local path (file or folder) to one single item in the catalog - by default as a top level item.
    The name of new item will be equal to the local filename.
    """
    remote_path = bpio.remotePath(name)
    key_alias = 'master' if not key_id else key_id.split('$')[0]
    if iter is None:
        iter = fs(key_alias=key_alias)
    if iterID is None:
        iterID = fsID(key_alias=key_alias)
    # make an ID for the filename
    newItemID = MakeID(iter)
    resultID = (parent_path_id.strip('/') + '/' + str(newItemID)).strip('/')
    typ = DIR if as_folder else FILE
    ii = FSItemInfo(name=remote_path, path_id=resultID, typ=typ, key_id=key_id)
    iter[ii.name()] = newItemID
    iterID[newItemID] = ii
    return resultID, ii, iter, iterID


#------------------------------------------------------------------------------


def SetFile(item, customer_idurl=None):
    """
    Put existing FSItemInfo ``item`` (for some single file) into the index.

    This is used when loading index from file. Should create all parent
    items in the index.

    Returns two boolean flags: success or not, modified or not
    """
    key_alias = item.key_alias()
    iter = fs(customer_idurl, key_alias)
    iterID = fsID(customer_idurl, key_alias)
    parts = item.path_id.lstrip('/').split('/')
    for j in range(len(parts)):
        part = parts[j]
        id = misc.ToInt(part, part)
        if j == len(parts) - 1:
            if item.name() not in iter:
                iter[item.name()] = id
                iterID[id] = item
                return True, True
            if item.pack_versions() == iterID[id].pack_versions():
                return True, False
            iterID[id] = item
            lg.warn('updated list of versions for %r' % item)
            return True, True
        found = False
        for name in iter.keys():
            if name == 0:
                continue
            if isinstance(iter[name], dict):
                if iter[name][0] == id:
                    iter = iter[name]
                    iterID = iterID[id]
                    found = True
                    break
                continue
        if not found:
            return False, False
    return False, False


def SetDir(item, customer_idurl=None):
    """
    Same, but ``item`` is a folder.
    """
    key_alias = item.key_alias()
    iter = fs(customer_idurl, key_alias)
    iterID = fsID(customer_idurl, key_alias)
    parts = item.path_id.lstrip('/').split('/')
    itemname = item.name()
    for j in range(len(parts)):
        part = parts[j]
        id = misc.ToInt(part, part)
        if j == len(parts) - 1:
            modified = False
            if itemname not in iter:
                iter[itemname] = {}
                modified = True
            if iter[itemname].get(0) != int(id):
                modified = True
            iter[itemname][0] = int(id)
            if id not in iterID:
                iterID[id] = {}
                modified = True
            cur_item = iterID[id].get(INFO_KEY)
            if not cur_item:
                modified = True
            iterID[id][INFO_KEY] = item
            return True, modified
        found = False
        for name in iter.keys():
            if name == 0:
                continue
            if isinstance(iter[name], int):
                continue
            if isinstance(iter[name], dict):
                if iter[name][0] == id:
                    iter = iter[name]
                    iterID = iterID[id]
                    found = True
                    break
                continue
            if strng.is_string(iter[name]):
                if iter[name] == itemname:
                    iter = iter[name]
                    iterID = iterID[id]
                    found = True
                    break
                continue
            raise Exception('wrong data type in the index')
        if not found:
            return False, False
    return False, False


#------------------------------------------------------------------------------


def WalkByPath(path, iter=None):
    """
    Search for ``path`` in the index - starting from root node if ``iter`` is None.

    Return None or tuple (iterator, ID).

      >>> backup_fs.WalkByPath('C:/Program Files/7-Zip/7z.exe')
      (3, '0/0/0/3')
      >>> backup_fs.WalkByPath('C:/Program Files/7-Zip/Lang/')
      ({0: 10, u'ru.txt': 1, u'en.ttt': 0}, '0/0/0/10')
    """
    if iter is None:
        iter = fs()
    ppath = bpio.remotePath(path)
    if ppath in iter:
        if isinstance(iter[ppath], int):
            return iter, str(iter[path])
        return iter[ppath], str(iter[ppath][0])
    if ppath == '' or ppath == '/':
        return iter, iter[0] if 0 in iter else ''
    path_id = ''
    parts = ppath.lstrip('/').split('/')
    for j in range(len(parts)):
        name = parts[j]
        if name not in iter:
            return None
        if isinstance(iter[name], dict):
            if 0 not in iter[name]:
                raise Exception('file or directory ID missed in the index')
            path_id += '/' + str(iter[name][0])
        elif isinstance(iter[name], int):
            if j != len(parts) - 1:
                return None
            path_id += '/' + str(iter[name])
        else:
            raise Exception('wrong data type in the index')
        if j == len(parts) - 1:
            return iter[name], path_id.lstrip('/')
        iter = iter[name]
    return None


def WalkByID(pathID, iterID=None):
    """
    Same, but search by ID:

        >>> backup_fs.WalkByID('0/0/0/10/1')
        (<FILE ru.txt 19107>, u'c:/Program Files/7-Zip/Lang/ru.txt')

    Both "walk" operations is working at O(log(n)) performance.
    """
    if iterID is None:
        iterID = fsID()
    if pathID is None:
        return None
    if pathID.strip() == '' or pathID.strip() == '/':
        return iterID, ''
    path = ''
    parts = pathID.strip('/').split('/')
    for j in range(len(parts)):
        part = parts[j]
        id = misc.ToInt(part, part)
        if id not in iterID:
            return None
        if isinstance(iterID[id], dict):
            if INFO_KEY not in iterID[id]:
                raise Exception('directory info missed in the index')
            name = iterID[id][INFO_KEY].name()
            if bpio.pathIsDriveLetter(name) or bpio.pathIsNetworkLocation(name):
                path += name
            else:
                path += '/' + name
        elif isinstance(iterID[id], FSItemInfo):
            if j != len(parts) - 1:
                return None
            path += '/' + iterID[id].name()
        else:
            raise Exception('wrong data type in the index')
        if j == len(parts) - 1:
            return iterID[id], ResolvePath(path)
        iterID = iterID[id]
    return None


#------------------------------------------------------------------------------


def DeleteByID(pathID, iter=None, iterID=None):
    """
    Delete item from index and return its path or None if not found.
    """
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    path = ''
    parts = pathID.strip('/').split('/')
    for j in range(len(parts)):
        part = parts[j]
        id = misc.ToInt(part, part)
        if id not in iterID:
            return None
        if isinstance(iterID[id], dict):
            if INFO_KEY not in iterID[id]:
                raise Exception('directory info missed in the index')
            name = iterID[id][INFO_KEY].name()
            if bpio.pathIsDriveLetter(name) or bpio.pathIsNetworkLocation(name):
                path += name
            else:
                path += '/' + name
        elif isinstance(iterID[id], FSItemInfo):
            name = iterID[id].name()
            path += '/' + name
            if j != len(parts) - 1:
                return None
        else:
            raise Exception('wrong data type in the index')
        if name not in iter:
            raise Exception('can not found target name in the index')
        if j == len(parts) - 1:
            iterID.pop(id)
            iter.pop(name)
            return path
        iterID = iterID[id]
        iter = iter[name]
    return None


def DeleteByPath(path, iter=None, iterID=None):
    """
    Delete given ``path`` from the index and return its ID.
    """
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    path_id = ''
    ppath = bpio.remotePath(path)
    parts = ppath.lstrip('/').split('/')
    if ppath in iter:
        path_id = iter[ppath]
        iter.pop(ppath)
        iterID.pop(path_id)
        return str(path_id)
    for j in range(len(parts)):
        name = parts[j]
        if name not in iter:
            return None
        if isinstance(iter[name], dict):
            if 0 not in iter[name]:
                raise Exception('directory ID missed in the index')
            id = iter[name][0]
            path_id += '/' + str(id)
        elif isinstance(iter[name], int):
            id = iter[name]
            path_id += '/' + str(id)
            if j != len(parts) - 1:
                return None
        else:
            raise Exception('wrong data type in the index')
        if id not in iterID:
            raise Exception('can not found target ID in the index')
        if j == len(parts) - 1:
            iter.pop(name)
            iterID.pop(id)
            return path_id.lstrip('/')
        iter = iter[name]
        iterID = iterID[id]
    return None


def DeleteBackupID(backupID, iterID=None):
    """
    Return backup from the index by its full ID.
    """
    keyAlias, customerGlobalID, remotePath, versionName = packetid.SplitBackupIDFull(backupID)
    if remotePath is None:
        lg.warn('%r has wrong format, remote path is not recognized' % backupID)
        return False
    customer_idurl = global_id.GlobalUserToIDURL(customerGlobalID)
    if iterID is None:
        iterID = fsID(customer_idurl, keyAlias)
    info = GetByID(remotePath, iterID=iterID)
    if info is None:
        lg.warn('not able to find file info for %r' % remotePath)
        return False
    if not info.has_version(versionName):
        lg.warn('%s do not have version %s' % (remotePath, versionName))
        return False
    info.delete_version(versionName)
    if _Debug:
        lg.args(_DebugLevel, backupID=backupID, versionName=versionName)
    return True


#------------------------------------------------------------------------------


def ToID(lookup_path, iter=None):
    """
    A wrapper for ``WalkByPath()`` method.
    """
    iter_and_id = WalkByPath(lookup_path, iter=iter)
    if iter_and_id is None:
        return None
    return iter_and_id[1]


def ToPath(pathID, iterID=None):
    """
    Get a full relative path from ``pathID``, this is a wrapper for
    ``WalkByID()`` method::

        >>> backup_fs.ToPath("/0/0/12/1")
        u'/home/veselin/Documents/somefile.txt'

    Return None if that ``pathID`` does not exist in the index.
    """
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return None
    return iter_and_path[1]


#------------------------------------------------------------------------------


def GetByID(pathID, iterID=None):
    """
    Return iterator to item with given ID, search in the index with
    ``WalkByID()``.
    """
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return None
    if isinstance(iter_and_path[0], dict):
        return iter_and_path[0][INFO_KEY]
    return iter_and_path[0]


def GetByPath(path, iter=None, iterID=None):
    """
    This calls ``ToID()`` first to get the ID and than use ``GetByID()`` to
    take the item.

    In fact we store items by ID, to have fast search by path we also
    keep opposite index. So we need to use second index to get the ID at
    first and than use the main index to get the item.
    """
    path_id = ToID(path, iter=iter)
    if not path_id:
        return None
    return GetByID(path_id, iterID=iterID)


def GetIteratorsByPath(path, iter=None, iterID=None):
    """
    Returns both iterators (iter, iterID) for given path, or None if not found.
    """
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return None
    iter_and_path = WalkByID(iter_and_id[1], iterID=iterID)
    if iter_and_path is None:
        return None
    return iter_and_id[1], iter_and_id[0], iter_and_path[0]


#------------------------------------------------------------------------------


def IsDir(path, iter=None):
    """
    Use ``WalkByPath()`` to check if the item with that ``path`` is directory.
    """
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return False
    iter, pathid = iter_and_id
    if pathid == '':
        return True
    if not isinstance(iter, dict):
        return False
    if 0 not in iter:
        raise Exception('directory ID missed in the index')
    return True


def IsFile(path, iter=None):
    """
    Same, but return True if this is a file.
    """
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return False
    iter, pathid = iter_and_id
    if isinstance(iter, int):
        return True
    if path in iter and isinstance(iter[path], int):
        return True
    return False


def IsDirID(pathID, iterID=None):
    """
    Return True if item with that ID is folder.
    """
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return False
    iterID, path = iter_and_path
    if isinstance(iterID, FSItemInfo):
        return False
    if INFO_KEY not in iterID:
        raise Exception('directory info missed in the index')
    return True


def IsFileID(pathID, iterID=None):
    """
    Return True if item with that ID is a file.
    """
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return False
    iterID, path = iter_and_path
    if not isinstance(iterID, FSItemInfo):
        return False
    return True


def Exists(path, iter=None):
    """
    Use ``WalkByPath()`` to check existence if that ``path``.
    """
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return False
    return True


def ExistsID(pathID, iterID=None):
    """
    Use ``WalkByPath()`` to check existence if that ``ID`` in the catalog.
    """
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return False
    return True


def ExistsBackupID(backupID, iterID=None):
    """
    Return True if backup with that ``backupID`` exist in the index.
    """
    keyAlias, customerGlobalID, remotePath, version = packetid.SplitBackupIDFull(backupID)
    if not remotePath:
        return False
    if iterID is None:
        iterID = fsID(global_id.GlobalUserToIDURL(customerGlobalID), keyAlias)
    iter_and_path = WalkByID(remotePath, iterID=iterID)
    if iter_and_path is None:
        return False
    return iter_and_path[0].has_version(version)


#------------------------------------------------------------------------------


def HasChilds(path, iter=None):
    """
    Return True if that item has some childs in the index.
    """
    if iter is None:
        iter = fs()
    if not path or path == '/':
        return len(iter) > 0
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return False
    iter, _ = iter_and_id
    if not isinstance(iter, dict):
        return False
    if 0 not in iter:
        raise Exception('directory ID missed in the index')
    return len(iter) > 0


def HasChildsID(pathID, iterID=None):
    """
    Same, but access to item by its ID.
    """
    if iterID is None:
        iterID = fsID()
    if not pathID:
        return len(iterID) > 0
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return False
    iterID, _ = iter_and_path
    if not isinstance(iterID, dict):
        return False
    if INFO_KEY not in iterID:
        raise Exception('directory info missed in the index')
    return len(iterID) > 0


#------------------------------------------------------------------------------


def ResolvePath(head, tail=''):
    """
    Smart join of path locations when read items from catalog.
    """
    if not head or head.strip().lstrip('/') == '':
        return tail.strip().lstrip('/')
    if not tail or tail.strip().lstrip('/') == '':
        return head.strip().lstrip('/') if head else ''
    return head.strip().lstrip('/') + '/' + tail.strip().lstrip('/')


def TraverseByID(callback, iterID=None, base_path_id=''):
    """
    Calls method ``callback(path_id, path, info)`` for every item in the index.
    """
    if iterID is None:
        iterID = fsID()

    def recursive_traverse(i, path_id, path, cb):
        name = None
        if path not in ['', '/']:
            path += '/'
        if isinstance(i, FSItemInfo):
            cb(path_id, ResolvePath(path), i)
            return
        if INFO_KEY in i:
            name = i[INFO_KEY].name()
            path += name
            cb(path_id, ResolvePath(path), i[INFO_KEY])
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                recursive_traverse(i[id], path_id + '/' + str(id) if path_id else str(id), path, cb)
            elif isinstance(i[id], FSItemInfo):
                cb(
                    (path_id + '/' + str(id)).lstrip('/') if path_id else str(id),  # pathID
                    ResolvePath(path, i[id].name()),  # remotePath
                    i[id],  # item
                )
            else:
                raise Exception('wrong item of type %r in the index: %r' % (type(i[id]), i[id]))

    startpth = '' if bpio.Windows() else '/'
    recursive_traverse(iterID, base_path_id, startpth, callback)


def TraverseByIDSorted(callback, iterID=None):
    """
    Same but sort file and folder names before traversing child nodes.
    """
    if iterID is None:
        iterID = fsID()

    def recursive_traverse_sorted(i, path_id, path, cb):
        name = None
        if path not in ['', '/']:
            path += '/'
        if isinstance(i, FSItemInfo):
            cb(path_id, ResolvePath(path), i, False)
            return
        if INFO_KEY in i:
            name = i[INFO_KEY].name()
            path += name
            cb(path_id, ResolvePath(path), i[INFO_KEY], len(i) > 1)
        dirs = []
        files = []
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                dirs.append((id, ResolvePath(path, i[id][INFO_KEY].name())))
            elif isinstance(i[id], FSItemInfo):
                files.append((id, ResolvePath(path, i[id].name())))
            else:
                raise Exception('wrong item type in the index')
        dirs.sort(key=lambda e: e[1])
        files.sort(key=lambda e: e[1])
        for id, pth in dirs:
            recursive_traverse_sorted(i[id], (path_id + '/' + str(id)).lstrip('/') if path_id else str(id), path, cb)
        for id, pth in files:
            cb((path_id + '/' + str(id)).lstrip('/') if path_id else str(id), ResolvePath(pth), i[id], False)
        del dirs
        del files

    startpth = '' if bpio.Windows() else '/'
    recursive_traverse_sorted(iterID, '', startpth, callback)


def TraverseChildsByID(callback, iterID=None):
    if iterID is None:
        iterID = fsID()

    def list_traverse(i, path_id, path, cb):
        name = None
        if path not in ['', '/']:
            path += '/'
        if isinstance(i, FSItemInfo):
            return
        if INFO_KEY in i:
            info = i[INFO_KEY]
            name = info.name()
            path += name
        dirs = []
        files = []
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                item_name = i[id][INFO_KEY].name()
                dirs.append((id, item_name, len(i[id]) - 1))
            elif isinstance(i[id], FSItemInfo):
                item_name = i[id].name()
                files.append((id, item_name))
            else:
                raise Exception('wrong item type in the index')
        dirs.sort(key=lambda e: e[1])
        files.sort(key=lambda e: e[1])
        for id, pth, num_childs in dirs:
            cb(DIR, ResolvePath(pth), (path_id + '/' + str(id)).lstrip('/') if path_id else str(id), i[id][INFO_KEY], num_childs)
        for id, pth in files:
            cb(FILE, ResolvePath(pth), (path_id + '/' + str(id)).lstrip('/') if path_id else str(id), i[id], False)
        del dirs
        del files

    startpth = '' if bpio.Windows() else '/'
    list_traverse(iterID, '', startpth, callback)


def IterateIDs(iterID=None):
    if iterID is None:
        iterID = fsID()

    def recursive_iterate(i, path_id, path):
        name = None
        if path not in ['', '/']:
            path += '/'
        if INFO_KEY in i:
            name = i[INFO_KEY].name()
            path += name
            yield path_id, ResolvePath(path), i[INFO_KEY]
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                for t in recursive_iterate(i[id], (path_id + '/' + str(id)).lstrip('/') if path_id else str(id), path):
                    yield t
            elif isinstance(i[id], FSItemInfo):
                yield (
                    (path_id + '/' + str(id)).lstrip('/') if path_id else str(id),  # pathID
                    ResolvePath(path, i[id].name()),  # remotePath
                    i[id],  # item
                )
            else:
                raise Exception('wrong item type in the index')

    startpth = '' if bpio.Windows() else '/'
    return recursive_iterate(iterID, '', startpth)


#------------------------------------------------------------------------------


def ExtractVersions(pathID, item_info, path_exist=None, customer_id=None, backup_info_callback=None):
    if not customer_id:
        customer_id = item_info.key_id or my_id.getGlobalID(key_alias='master')
    item_size = 0
    item_time = 0
    versions = []
    for version, version_info in item_info.versions.items():
        backupID = packetid.MakeBackupID(customer_id, pathID, version)
        version_time = misc.TimeFromBackupID(version)
        if version_time and version_time > item_time:
            item_time = version_time
        version_size = version_info[1]
        if version_size > 0:
            item_size += version_size
        if packetid.IsCanonicalVersion(version):
            # 0 1234 56 78 9  11 13 15
            # F 2013 11 20 05 38 03 PM
            b = version
            version_label = '%s-%s-%s %s:%s:%s %s' % (
                b[1:5],
                b[5:7],
                b[7:9],
                b[9:11],
                b[11:13],
                b[13:15],
                b[15:17],
            )
        else:
            version_label = backupID
        backup_info_dict = {
            'backup_id': backupID,
            'label': version_label,
            'time': version_time,
            'size': version_size,
        }
        if backup_info_callback:
            backup_info_dict.update(backup_info_callback(backupID, item_info, item_info.name(), path_exist))
        versions.append(backup_info_dict)
    item_time = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(item_time)) if item_time else ''
    return item_size, item_time, versions


#------------------------------------------------------------------------------


def ListAllBackupIDs(customer_idurl=None, sorted=False, reverse=False):
    """
    Traverse all index and list all backup IDs for all of the customers.
    """
    lst = []

    def visitor(path_id, _, info):
        for version in info.list_versions(sorted, reverse):
            lst.append(packetid.MakeBackupID(info.key_id, path_id.lstrip('/'), version))

    if customer_idurl is None:
        for customer_idurl in known_customers():
            for key_alias in known_keys_aliases(customer_idurl):
                TraverseByID(visitor, iterID=fsID(customer_idurl, key_alias))
    else:
        for key_alias in known_keys_aliases(customer_idurl):
            TraverseByID(visitor, iterID=fsID(customer_idurl, key_alias))
    return lst


def ListAllBackupIDsFull(sorted=False, reverse=False, iterID=None, base_path_id=None):
    """
    Same, but also return items info.
    """
    if iterID is None:
        iterID = fsID()
    if base_path_id is None:
        if INFO_KEY in iterID:
            base_path_id = iterID[INFO_KEY].path_id
    lst = []

    def visitor(path_id, path, info):
        for version in info.list_versions(sorted=sorted, reverse=reverse):
            backupID = packetid.MakeBackupID(info.key_id, path_id.lstrip('/'), version)
            lst.append((info.name(), backupID, info.get_version_info(version), ResolvePath(path), info))

    TraverseByID(visitor, iterID=iterID, base_path_id=base_path_id)
    return lst


def ListChildsByPath(path, recursive=False, iter=None, iterID=None, backup_info_callback=None):
    """
    List all items at given ``path`` and return data as a list of dict objects.
    Return string with error message if operation failed.
    """
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    if path == '/':
        path = ''
    path = bpio.remotePath(path)
    iter_and_id = WalkByPath(path, iter=iter)
    if iter_and_id is None:
        return 'path "%s" not found' % path
    iter, pathID = iter_and_id
    iter_and_path = WalkByID(pathID, iterID=iterID)
    if iter_and_path is None:
        return 'item "%s" exist, but not path "%s" not found, catalog index is not consistent' % (pathID, path)
    iterID, path_exist = iter_and_path
    if path != path_exist:
        return 'item "%s" exist, but path "%s" is not valid, catalog index is not consistent' % (path_exist, path)
    if isinstance(iterID, FSItemInfo):
        return 'path "%s" is a file' % path
    result = []
    sub_dirs = []

    def visitor(item_type, item_name, item_path_id, item_info, num_childs):
        item_id = (pathID + '/' + item_path_id).strip('/')
        item_size, item_time, versions = ExtractVersions(item_id, item_info, path_exist, backup_info_callback=backup_info_callback)
        i = {
            'type': item_type,
            'name': item_info.name(),
            'path': ResolvePath(path, item_info.name()),
            'path_id': item_id,
            'total_size': item_size,
            'latest': item_time,
            'childs': num_childs,
            'item': item_info.serialize(to_json=True),
            'versions': versions,
        }
        result.append(i)
        if item_type == DIR:
            sub_dirs.append(i)

    TraverseChildsByID(visitor, iterID)

    if recursive:
        for sub_dir in sub_dirs:
            sub_lookup = ListChildsByPath(sub_dir['path'], recursive=False, iter=iter, iterID=iterID, backup_info_callback=backup_info_callback)
            if not isinstance(sub_lookup, list):
                return sub_lookup
            result.extend(sub_lookup)
    return result


#------------------------------------------------------------------------------


def MakeLocalDir(basedir, backupID):
    """
    This creates a local folder for that given ``backupID`` - this is a relative path,
    a ``basedir`` must be a root folder.

    For example::
      MakeLocalDir("c:/temp", "alice@p2p.net:0/1/2/3/F123456")

    should create a folder with such absolute path::
      c:/temp/alice@p2p.net/0/1/2/3/F123456

    Do some checking and call built-in method ``os.makedirs()``.
    """
    if not bpio.pathIsDir(basedir):
        raise Exception('directory not exist: %s' % basedir)
    customerGlobalID, remotePath = packetid.SplitPacketID(backupID)
    if not customerGlobalID:
        customerGlobalID = my_id.getGlobalID()
    path = os.path.join(basedir, customerGlobalID, remotePath)
    if os.path.exists(path):
        if not bpio.pathIsDir(path):
            raise Exception('can not create directory %s' % path)
    else:
        os.makedirs(path)
    return path


def DeleteLocalDir(basedir, pathID):
    """
    Remove local sub folder at given ``basedir`` root path.
    """
    if not bpio.pathIsDir(basedir):
        raise Exception('directory not exist: %s' % basedir)
    customer, pth = packetid.SplitPacketID(pathID)
    path = os.path.join(basedir, customer, pth)
    if not os.path.exists(path):
        return
    if not bpio.pathIsDir(path):
        raise Exception('path %s is not a directory' % path)
    bpio.rmdir_recursive(path, ignore_errors=True)


def DeleteLocalBackup(basedir, backupID):
    """
    Remove local files for that backup.
    """
    count_and_size = [0, 0]
    if not bpio.pathIsDir(basedir):
        raise Exception('directory not exist: %s' % basedir)
    customer, pth = packetid.SplitPacketID(backupID)
    backupDir = os.path.join(basedir, customer, pth)
    if not bpio.pathExist(backupDir):
        return count_and_size[0], count_and_size[1]
    if not bpio.pathIsDir(backupDir):
        raise Exception('path %s is not a directory' % backupDir)

    def visitor(fullpath):
        if os.path.isfile(fullpath):
            try:
                count_and_size[1] += os.path.getsize(fullpath)
                count_and_size[0] += 1
            except:
                pass
        return True

    counter = bpio.rmdir_recursive(backupDir, ignore_errors=True, pre_callback=visitor)
    if _Debug:
        lg.args(_DebugLevel, backupDir=backupDir, counter=counter)
    return count_and_size[0], count_and_size[1]


#------------------------------------------------------------------------------


def Scan(basedir=None, customer_idurl=None, key_alias='master'):
    """
    Walk all items in the index and check if local files and folders with same
    names exists.

    Parameter ``basedir`` is a root path of that structure, default is
    ``lib.settings.getLocalBackupsDir()``. Also calculate size of the
    files.
    """
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    if basedir is None:
        basedir = settings.getLocalBackupsDir()
    iterID = fsID(customer_idurl, key_alias=key_alias)
    summ = [
        0,
        0,
    ]

    def visitor(path_id, path, info):
        info.read_stats(path)
        if info.exist():
            summ[0] += info.size
        k_alias = key_alias
        if info.key_id:
            k_alias = packetid.KeyAlias(info.key_id)
        customer_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias=k_alias)
        versions_path = bpio.portablePath(os.path.join(basedir, customer_id, path_id))
        summ[1] += info.read_versions(versions_path)

    TraverseByID(visitor, iterID=iterID)
    return summ[0], summ[1]


def ScanID(pathID, basedir=None, customer_idurl=None, key_alias='master'):
    """
    Same as `Scan`, but check only single item in the index.
    """
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    if basedir is None:
        basedir = settings.getLocalBackupsDir()
    iter_and_path = WalkByID(pathID, iterID=fs(customer_idurl, key_alias=key_alias))
    if not iter_and_path:
        return
    itr, path = iter_and_path
    if isinstance(iter, dict):
        if INFO_KEY not in iter:
            return
        itr = iter[INFO_KEY]
    k_alias = key_alias
    if itr and itr.key_id:
        k_alias = packetid.KeyAlias(itr.key_id)
    customer_id = global_id.MakeGlobalID(idurl=customer_idurl, key_alias=k_alias)
    itr.read_stats(path)
    itr.read_versions(bpio.portablePath(os.path.join(basedir, customer_id)))


def Calculate(iterID=None):
    """
    Scan all items in the index and calculate folder and backups sizes.
    """
    global _SizeFiles
    global _SizeFolders
    global _SizeBackups
    global _ItemsCount
    global _FilesCount
    _ItemsCount = 0
    _FilesCount = 0
    _DirsCount = 0
    _SizeFiles = 0
    _SizeFolders = 0
    _SizeBackups = 0

    def recursive_calculate(i):
        global _SizeFiles
        global _SizeFolders
        global _SizeBackups
        global _ItemsCount
        global _FilesCount
        global _DirsCount
        folder_size = 0
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], FSItemInfo):
                if i[id].exist():
                    folder_size += i[id].size
                if i[id].type == FILE:
                    _FilesCount += 1
                    if i[id].exist():
                        _SizeFiles += i[id].size
                if i[id].type == DIR:
                    _DirsCount += 1
                    if i[id].exist():
                        _SizeFolders += i[id].size
                for version in i[id].list_versions():
                    versionSize = i[id].get_version_info(version)[1]
                    if versionSize > 0:
                        _SizeBackups += versionSize
                _ItemsCount += 1
            elif isinstance(i[id], dict):
                sub_folder_size = recursive_calculate(i[id])
                if sub_folder_size != -1:
                    folder_size += sub_folder_size
            else:
                raise Exception('wrong item type in the index')
        if INFO_KEY in i:
            i[INFO_KEY].size = folder_size
            if i[INFO_KEY].type == FILE:
                _FilesCount += 1
                if i[INFO_KEY].exist():
                    _SizeFiles += i[INFO_KEY].size
            if i[INFO_KEY].type == DIR:
                _DirsCount += 1
                if i[INFO_KEY].exist():
                    _SizeFolders += i[INFO_KEY].size
            for version in i[INFO_KEY].list_versions():
                versionSize = i[INFO_KEY].get_version_info(version)[1]
                if versionSize > 0:
                    _SizeBackups += versionSize
            _ItemsCount += 1
        return folder_size

    if iterID is None:
        iterID = fsID()
    ret = recursive_calculate(iterID)
    if _Debug:
        lg.out(_DebugLevel, 'backup_fs.Calculate %d %d %d %d' % (_ItemsCount, _FilesCount, _SizeFiles, _SizeBackups))
    return ret


#------------------------------------------------------------------------------


def Clear(customer_idurl, key_alias=None):
    """
    Erase all items in the index for given customer and key alias and also forget the latest revision.
    """
    fs(customer_idurl, key_alias).clear()
    fsID(customer_idurl, key_alias).clear()
    # forget(customer_idurl, key_alias)


def ClearAllIndexes():
    global _FileSystemIndexByID
    global _FileSystemIndexByName
    global _RevisionNumber
    _FileSystemIndexByID.clear()
    _FileSystemIndexByName.clear()
    _RevisionNumber.clear()


#------------------------------------------------------------------------------


def SerializeIndex(customer_idurl, key_alias=None, encoding='utf-8', filter_cb=None):
    """
    Use this to write index to the local file.
    """
    cnt = [0]
    result = {}

    def cb(path_id, path, info, k_alias):
        if filter_cb is not None:
            if not filter_cb(path_id, path, info):
                return
        result[k_alias]['items'].append(info.serialize(encoding=encoding, to_json=True))
        cnt[0] += 1

    key_aliases = []
    if key_alias:
        key_aliases.append(key_alias)
    else:
        key_aliases.extend(known_keys_aliases(customer_idurl))
    for k_alias in key_aliases:
        if k_alias not in result:
            result[k_alias] = {}
        if 'items' not in result[k_alias]:
            result[k_alias]['items'] = []
        TraverseByID(
            callback=lambda path_id, path, info: cb(path_id, path, info, k_alias),
            iterID=fsID(customer_idurl, k_alias),
        )
    if _Debug:
        lg.dbg(_DebugLevel, 'done with %d indexed files of %d aliases for %r' % (cnt[0], len(key_aliases), customer_idurl))
    return result


def UnserializeIndex(json_data, customer_idurl=None, new_revision=None, deleted_path_ids=[], decoding='utf-8'):
    """
    Read index from ``StringIO`` object.
    """
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    total_count = 0
    total_modified_count = 0
    updated_keys = []
    for key_alias in json_data.keys():
        if new_revision is not None:
            cur_revision = revision(customer_idurl, key_alias)
            if cur_revision >= new_revision:
                if _Debug:
                    lg.dbg(_DebugLevel, 'ignore items for %r with alias %r because current revision is up to date: %d >= %d' % (customer_idurl, key_alias, cur_revision, new_revision))
                continue
        count = 0
        count_modified = 0
        modified_items = set()
        known_items = set()
        new_files = []
        to_be_removed_items = set()
        to_be_removed_items.update(deleted_path_ids)
        for json_item in json_data[key_alias]['items']:
            item = FSItemInfo()
            item.unserialize(json_item, decoding=decoding, from_json=True)
            if item.path_id in deleted_path_ids:
                continue
            known_items.add(item.path_id)
            if item.type == FILE:
                success, modified = SetFile(item, customer_idurl=customer_idurl)
                if not success:
                    lg.warn('Can not put FILE item into the tree: %s' % str(item))
                    raise ValueError('Can not put FILE item into the tree: %s' % str(item))
                count += 1
                if modified:
                    count_modified += 1
                    modified_items.add(item.path_id)
                    new_files.append(item)
            elif item.type == DIR:
                success, modified = SetDir(item, customer_idurl=customer_idurl)
                if not success:
                    lg.warn('Can not put DIR item into the tree: %s' % str(item))
                    raise ValueError('Can not put DIR item into the tree: %s' % str(item))
                count += 1
                if modified:
                    count_modified += 1
                    modified_items.add(item.path_id)
            else:
                raise ValueError('Incorrect entry type')

        if _Debug:
            lg.args(_DebugLevel, c=customer_idurl, k=key_alias, new_files=len(new_files))

        def _one_item(path_id, path, info):
            if path_id not in known_items:
                if path_id != settings.BackupIndexFileName():
                    to_be_removed_items.add(path_id)

        TraverseByID(_one_item, iterID=fsID(customer_idurl, key_alias))
        if _Debug:
            lg.dbg(_DebugLevel, 'from %d known items %d were modified and %d items marked to be removed' % (len(known_items), len(modified_items), len(to_be_removed_items)))
        for path_id in to_be_removed_items:
            deleted_info = {}
            if key_alias.startswith('share_'):
                deleted_iter_and_path = WalkByID(path_id, iterID=fsID(customer_idurl, key_alias))
                if deleted_iter_and_path:
                    deleted_file_item, deleted_file_path = deleted_iter_and_path
                    full_glob_id = global_id.MakeGlobalID(idurl=customer_idurl, path=path_id, key_alias=key_alias)
                    full_remote_path = global_id.MakeGlobalID(idurl=customer_idurl, path=deleted_file_path, key_alias=key_alias)
                    deleted_info = dict(
                        global_id=full_glob_id,
                        remote_path=full_remote_path,
                        size=max(0, deleted_file_item.size),
                        type=TYPES.get(deleted_file_item.type, 'unknown').lower(),
                        customer=customer_idurl.to_id(),
                        versions=[dict(backup_id=v) for v in deleted_file_item.versions.keys()],
                    )
            DeleteByID(path_id, iter=fs(customer_idurl, key_alias), iterID=fsID(customer_idurl, key_alias))
            count_modified += 1
            if deleted_info:
                snapshot = dict(
                    global_id=deleted_info['global_id'],
                    remote_path=deleted_info['remote_path'],
                    size=deleted_info['size'],
                    type=deleted_info['type'],
                    customer=deleted_info['customer'],
                    versions=deleted_info['versions'],
                )
                listeners.push_snapshot('shared_file', snap_id=full_glob_id, deleted=True, data=snapshot)
            if driver.is_on('service_shared_data'):
                from bitdust.access import shared_access_coordinator
                shared_access_coordinator.on_file_deleted(customer_idurl, key_alias, path_id)

        total_count += count
        total_modified_count += count_modified
        if new_revision is not None and new_revision > cur_revision:
            old_rev = None
            new_rev = None
            old_rev, new_rev = commit(new_revision_number=new_revision, customer_idurl=customer_idurl, key_alias=key_alias)
            Scan(customer_idurl=customer_idurl, key_alias=key_alias)
            updated_keys.append(key_alias)
            if key_alias.startswith('share_'):
                for new_file_item in new_files:
                    if new_file_item.path_id == settings.BackupIndexFileName():
                        continue
                    new_file_path = ToPath(new_file_item.path_id, iterID=fsID(customer_idurl, key_alias))
                    if new_file_path:
                        full_glob_id = global_id.MakeGlobalID(idurl=customer_idurl, path=new_file_item.path_id, key_alias=key_alias)
                        full_remote_path = global_id.MakeGlobalID(idurl=customer_idurl, path=new_file_path, key_alias=key_alias)
                        snapshot = dict(
                            global_id=full_glob_id,
                            remote_path=full_remote_path,
                            size=max(0, new_file_item.size),
                            type=TYPES.get(new_file_item.type, 'unknown').lower(),
                            customer=customer_idurl.to_id(),
                            versions=[dict(backup_id=v) for v in new_file_item.versions.keys()],
                        )
                        listeners.push_snapshot('shared_file', snap_id=full_glob_id, data=snapshot)
            if _Debug:
                lg.args(_DebugLevel, count=count, modified=count_modified, c=customer_idurl, k=key_alias, old_rev=old_rev, new_rev=new_rev)
    if _Debug:
        lg.dbg(_DebugLevel, 'done with %d total items and %d modified, loaded data for %d keys' % (total_count, total_modified_count, len(updated_keys)))
    return total_count, total_modified_count, updated_keys


#------------------------------------------------------------------------------


def SaveIndex(customer_idurl=None, key_alias='master', encoding='utf-8'):
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    customer_id = customer_idurl.to_id()
    index_file_path = settings.BackupIndexFilePath(customer_idurl, key_alias)
    if not os.path.isdir(os.path.dirname(index_file_path)):
        os.makedirs(os.path.dirname(index_file_path))
    json_data = {}
    json_data[customer_id] = SerializeIndex(
        customer_idurl=customer_idurl,
        key_alias=key_alias,
        encoding=encoding,
    )
    rev = revision(customer_idurl, key_alias)
    src = '%d\n' % rev
    src += jsn.dumps(
        json_data,
        indent=1,
        separators=(',', ':'),
        encoding=encoding,
    )
    if _Debug:
        lg.args(_DebugLevel, rev=rev, c=customer_id, k=key_alias, sz=len(src), path=index_file_path)
    return bpio.WriteTextFile(index_file_path, src)


def ReadIndex(text_data, new_revision=None, deleted_path_ids=[], encoding='utf-8'):
    total_count = 0
    total_modified_count = 0
    updated_customers_keys = []
    try:
        json_data = jsn.loads(text_data, encoding=encoding)
    except:
        lg.exc()
        return 0, []
    if _Debug:
        lg.args(_DebugLevel, new_revision=new_revision, sz=len(text_data), deleted=deleted_path_ids)
    if not json_data:
        return 0, []
    for customer_id in json_data.keys():
        customer_idurl = global_id.GlobalUserToIDURL(customer_id)
        if not id_url.is_cached(customer_idurl):
            lg.warn('identity %r is not yet cached, skip reading related catalog items' % customer_idurl)
            identitycache.immediatelyCaching(customer_idurl, try_other_sources=False, ignore_errors=True)
            continue
        try:
            count, modified_count, updated_keys = UnserializeIndex(
                json_data[customer_id],
                customer_idurl=customer_idurl,
                new_revision=new_revision,
                deleted_path_ids=deleted_path_ids,
                decoding=encoding,
            )
        except:
            lg.exc()
            continue
        total_count += count
        total_modified_count += modified_count
        if updated_keys:
            for key_alias in updated_keys:
                Calculate(iterID=fsID(customer_idurl, key_alias))
                updated_customers_keys.append((customer_idurl, key_alias))
    if _Debug:
        lg.out(_DebugLevel, 'backup_fs.ReadIndex %d items loaded for %d keys' % (total_count, len(updated_customers_keys)))
    return total_count, updated_customers_keys


def LoadIndex(index_file_path):
    src = bpio.ReadTextFile(index_file_path)
    if not src:
        lg.err('failed reading file %s' % index_file_path)
        return False
    inpt = StringIO(src)
    try:
        new_revision = int(inpt.readline().rstrip('\n'))
    except:
        lg.exc()
        return False
    raw_data = inpt.read()
    inpt.close()
    count, _ = ReadIndex(raw_data, new_revision=new_revision)
    if not count:
        return False
    return count


def LoadAllIndexes():
    index_dir_path = os.path.join(settings.ServiceDir('service_backups'), 'index')
    if not os.path.isdir(index_dir_path):
        os.makedirs(index_dir_path)
    for key_id in os.listdir(index_dir_path):
        if my_keys.latest_key_id(key_id) != key_id:
            lg.warn('ignore old index file for rotated identity: %r' % key_id)
            continue
        index_file_path = os.path.join(index_dir_path, key_id)
        LoadIndex(index_file_path)


#------------------------------------------------------------------------------


def populate_private_files():
    ret = api.files_list(remote_path='', key_id=my_id.getGlobalID(key_alias='master'))
    if ret['status'] != 'OK':
        return
    lst = ret['result']
    for itm in lst:
        if itm['path'] == 'index':
            continue
        snapshot = dict(
            global_id=itm['global_id'],
            remote_path=itm['remote_path'],
            size=itm['size'],
            type=itm['type'],
            customer=itm['customer'],
            versions=[dict(backup_id=v['backup_id'].split('/')[-1]) for v in itm['versions']],
        )
        listeners.push_snapshot('private_file', snap_id=itm['global_id'], data=snapshot)


def populate_shared_files(key_id=None):
    lst = []
    if key_id:
        ret = api.files_list(remote_path='', key_id=key_id)
        if ret['status'] != 'OK':
            return
        lst = ret['result']
    else:
        ret = api.shares_list()
        if ret['status'] != 'OK':
            return
        for one_share in ret['result']:
            ret = api.files_list(remote_path='', key_id=one_share['key_id'])
            if ret['status'] != 'OK':
                return
            lst.extend(ret['result'])
    for itm in lst:
        if itm['path'] == 'index':
            continue
        snapshot = dict(
            global_id=itm['global_id'],
            remote_path=itm['remote_path'],
            size=itm['size'],
            type=itm['type'],
            customer=itm['customer'],
            versions=[dict(backup_id=v['backup_id'].split('/')[-1]) for v in itm['versions']],
        )
        listeners.push_snapshot('shared_file', snap_id=itm['global_id'], data=snapshot)


#------------------------------------------------------------------------------


def _test():
    """
    For tests.
    """
    import pprint
    settings.init()
    filepath = settings.BackupIndexFilePath()
    # print filepath
    src = bpio.ReadTextFile(filepath)
    inpt = StringIO(src)
    inpt.readline()
    json_data = json.loads(inpt.read())
    inpt.close()
    for customer_id in json_data.keys():
        count, modified_count, updated_keys = UnserializeIndex(json_data[customer_id])
        print(customer_id, count)
    Scan()
    Calculate()

    # pprint.pprint(fs())
    # print
    # pprint.pprint(fsID())
    # print

    #     print AddDir('dir1/dir2')
    #     ii = GetIteratorsByPath('dir1')
    #     print ii
    #     print PutItem('fff', as_folder=False, iter=ii[0], iterID=ii[1])
    #
    #     print IsDir('dir1')
    #     print IsFile('dir2/fff')

    print('------------')
    pprint.pprint(fs(key_alias='share_f17b49966dfe85320ac5e7d579d0047c'))
    pprint.pprint(ListChildsByPath(
        path='',
        recursive=True,
        iter=fs(key_alias='share_f17b49966dfe85320ac5e7d579d0047c'),
        iterID=fsID(key_alias='share_f17b49966dfe85320ac5e7d579d0047c'),
    ))
    # print()
    # pprint.pprint(fsID())
    print('------------')

    # print(HasChilds('', iter=fs(customer_idurl)))
    # pprint.pprint([i[1] for i in ListAllBackupIDsFull()])
    # pprint.pprint([i[1] for i in ListAllBackupIDsFull(iterID=WalkByID('0')[0])])

    settings.shutdown()


#     for i in range(10000):
#         r = AddFile('file' + str(i))
#         print r[0], len(fs())
#         # pprint.pprint(fs())

# PutItem('dir4', as_folder=True)

#     print ListByPathAdvanced('TestKey2')
#     return

#     for i in ListByPathAdvanced('TestKey2'):
#         print i # , i[1] # , WalkByPath(i[1])[1]
#         print

#     parent_path = os.path.dirname(bpio.portablePath(unicode('/some/remote/path')))
#     print IsFile(sys.argv[1])
#     if iter_and_iterID is None:
#         _, parent_iter, parent_iterID = AddDir(parent_path, read_stats=False)
#     else:
#         parent_iter, parent_iterID = iter_and_iterID
#     print MapPath('/tmp/com.apple.launchd.KlOCuOQw9M', iter=parent_iter, iterID=parent_iterID)

#     print AddDir('/this/folder/not/exist/', read_stats=True, keyID='KKKKKKK')
#     print AddDir('/this/folder/not/exist/subdir')

#     pprint.pprint(fs())
#     pprint.pprint(fsID())

# print GetByPath('')

# print WalkByID('8/0/0')

# print AppendFile('new', '/Users/veselin/Pictures')

# item = FSItemInfo('new', '/Users/new', FILE)
# SetFile(item)
# print ToID('dir6/ilhan.jpg')
# print WalkByPath('dir6')

# print Exists('/asd')
# print IsDir('/asd')
# MapPath("/Users/veselin/Pictures/fotosess/Thumbs.db")
# pprint.pprint(fs())
# pprint.pprint(fsID())
# for pathID, localPath, item in IterateIDs():
#     sz = diskspace.MakeStringFromBytes(item.size) if item.exist() else ''
#     print '  %s %s %s' % (pathID.ljust(27), localPath.ljust(70), sz.ljust(9))

# pprint.pprint(ListRootItems())
# pprint.pprint(ListAllBackupIDs())
# pprint.pprint(ListChildsByPath((sys.argv[1])))

# print ListByPathAdvanced("")
# pth = '~/Downloads/test/asd'
# ppth = bpio.portablePath(unicode(pth))
# print ppth
# print ToID(ppth)

#    bpio.init()
#    for path in sys.argv[1:]:
#        print path, AddLocalPath(path, True)
#
#    for pathID, localPath, itemInfo in IterateIDs():
#        print pathID, localPath, itemInfo
#
#    print IsDirID('0/0/0')

# pprint.pprint(fs())
# pprint.pprint(fsID())
# s = Serialize()
# print s
#    open('index', 'w').write(s)
#    for id, path in ListByID(''):
#        pprint.pprint(ListByID(id))
#    fs().clear()
#    fsID().clear()
#    inp = StringIO(s)
#    Unserialize(inp)
#    inp.close()
#    pprint.pprint(fs())
#    pprint.pprint(fsID())

# print GetPath('/0/0/0/1/31/12')
# print GetID('C:/.Veselin/diplom/Social/Editor/type 5.vsd.aut')
# print WalkByPath('C:/.Veselin/diplom/Social/Editor/type 5.vsd.aut')
# print WalkByID('/0/0/0/1/31/12')
# print IsDir('C:')
# DeleteByID('/0')
# print ListByID('')
# print 'NEW C:/.Veselin/diplom/', AddLocalPath('C:/.Veselin/diplom/')
# print ListByID('')
# pprint.pprint(ListByPath('c:\.Veselin/diplom/'))
# print DeleteByID('0/0/0/0')
# print DeleteByPath('C:/.Veselin/diplom/Social')
# pprint.pprint(fs())
# pprint.pprint(ListByPath('C:/.Veselin/diplom/Social/Editor'))

# iter = fs()['c:']['work']['modnamama']['_katalog']
# print map(lambda k: iter[k][0] if isinstance(iter, dict) else '', iter.keys())
# pprint.pprint( ListLocalFolder(sys.argv[1]) )
# print ListLocalFolder(sys.argv[1])

if __name__ == '__main__':
    _test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import re
import tokenize
import token
import uuid
from six.moves import range


class IndexCreatorException(Exception):

    def __init__(self, ex, line=None):
        self.ex = ex
        self.line = line

    def __str__(self):
        if self.line:
            return repr(self.ex + "(in line: %d)" % self.line)
        return repr(self.ex)


class IndexCreatorFunctionException(IndexCreatorException):
    pass


class IndexCreatorValueException(IndexCreatorException):
    pass


class Parser(object):

    def __init__(self):
        pass

    def parse(self, data, name=None):
        if not name:
            self.name = "_" + uuid.uuid4().hex
        else:
            self.name = name

        self.ind = 0
        self.stage = 0
        self.logic = ['and', 'or', 'in']
        self.logic2 = ['&', '|']
        self.allowed_props = {'TreeBasedIndex': ['type', 'name', 'key_format', 'node_capacity', 'pointer_format', 'meta_format'],
                              'HashIndex': ['type', 'name', 'key_format', 'hash_lim', 'entry_line_format'],
                              'MultiHashIndex': ['type', 'name', 'key_format', 'hash_lim', 'entry_line_format'],
                              'MultiTreeBasedIndex': ['type', 'name', 'key_format', 'node_capacity', 'pointer_format', 'meta_format']
                              }
        self.funcs = {'md5': (['md5'], ['.digest()']),
                      'len': (['len'], []),
                      'str': (['str'], []),
                      'fix_r': (['self.fix_r'], []),
                      'prefix': (['self.prefix'], []),
                      'infix': (['self.infix'], []),
                      'suffix': (['self.suffix'], [])
                      }
        self.handle_int_imports = {'infix': "from itertools import izip\n"}

        self.funcs_with_body = {'fix_r':
                                ("""    def fix_r(self,s,l):
        e = len(s)
        if e == l:
            return s
        elif e > l:
            return s[:l]
        else:
            return s.rjust(l,'_')\n""", False),
                                'prefix':
                                ("""    def prefix(self,s,m,l,f):
        t = len(s)
        if m < 1:
            m = 1
        o = set()
        if t > l:
            s = s[:l]
            t = l
        while m <= t:
            o.add(s.rjust(f,'_'))
            s = s[:-1]
            t -= 1
        return o\n""", False),
                                'suffix':
                                ("""    def suffix(self,s,m,l,f):
        t = len(s)
        if m < 1:
            m = 1
        o = set()
        if t > l:
            s = s[t-l:]
            t = len(s)
        while m <= t:
            o.add(s.rjust(f,'_'))
            s = s[1:]
            t -= 1
        return o\n""", False),
                                'infix':
                                ("""    def infix(self,s,m,l,f):
        t = len(s)
        o = set()
        for x in xrange(m - 1, l):
            t = (s, )
            for y in xrange(0, x):
                t += (s[y + 1:],)
            o.update(set(''.join(x).rjust(f, '_').lower() for x in izip(*t)))
        return o\n""", False)}
        self.none = ['None', 'none', 'null']
        self.props_assign = ['=', ':']
        self.all_adj_num_comp = {token.NUMBER: (
            token.NUMBER, token.NAME, '-', '('),
            token.NAME: (token.NUMBER, token.NAME, '-', '('),
            ')': (token.NUMBER, token.NAME, '-', '(')
        }

        self.all_adj_num_op = {token.NUMBER: (token.NUMBER, token.NAME, '('),
                               token.NAME: (token.NUMBER, token.NAME, '('),
                               ')': (token.NUMBER, token.NAME, '(')
                               }
        self.allowed_adjacent = {
            "<=": self.all_adj_num_comp,
            ">=": self.all_adj_num_comp,
            ">": self.all_adj_num_comp,
            "<": self.all_adj_num_comp,

            "==": {token.NUMBER: (token.NUMBER, token.NAME, '('),
                   token.NAME: (token.NUMBER, token.NAME, token.STRING, '('),
                   token.STRING: (token.NAME, token.STRING, '('),
                   ')': (token.NUMBER, token.NAME, token.STRING, '('),
                   ']': (token.NUMBER, token.NAME, token.STRING, '(')
                   },

            "+": {token.NUMBER: (token.NUMBER, token.NAME, '('),
                  token.NAME: (token.NUMBER, token.NAME, token.STRING, '('),
                  token.STRING: (token.NAME, token.STRING, '('),
                  ')': (token.NUMBER, token.NAME, token.STRING, '('),
                  ']': (token.NUMBER, token.NAME, token.STRING, '(')
                  },

            "-": {token.NUMBER: (token.NUMBER, token.NAME, '('),
                  token.NAME: (token.NUMBER, token.NAME, '('),
                  ')': (token.NUMBER, token.NAME, '('),
                  '<': (token.NUMBER, token.NAME, '('),
                  '>': (token.NUMBER, token.NAME, '('),
                  '<=': (token.NUMBER, token.NAME, '('),
                  '>=': (token.NUMBER, token.NAME, '('),
                  '==': (token.NUMBER, token.NAME, '('),
                  ']': (token.NUMBER, token.NAME, '(')
                  },
            "*": self.all_adj_num_op,
            "/": self.all_adj_num_op,
            "%": self.all_adj_num_op,
            ",": {token.NUMBER: (token.NUMBER, token.NAME, token.STRING, '{', '[', '('),
                  token.NAME: (token.NUMBER, token.NAME, token.STRING, '(', '{', '['),
                  token.STRING: (token.NAME, token.STRING, token.NUMBER, '(', '{', '['),
                  ')': (token.NUMBER, token.NAME, token.STRING, '(', '{', '['),
                  ']': (token.NUMBER, token.NAME, token.STRING, '(', '{', '['),
                  '}': (token.NUMBER, token.NAME, token.STRING, '(', '{', '[')
                  }
        }

        def is_num(s):
            m = re.search('[^0-9*()+\-\s/]+', s)
            return not m

        def is_string(s):
            m = re.search('\s*(?P<a>[\'\"]+).*?(?P=a)\s*', s)
            return m
        data = re.split('make_key_value\:', data)

        if len(data) < 2:
            raise IndexCreatorFunctionException(
                "Couldn't find a definition of make_key_value function!\n")

        spl1 = re.split('make_key\:', data[0])
        spl2 = re.split('make_key\:', data[1])

        self.funcs_rev = False

        if len(spl1) > 1:
            data = [spl1[0]] + [data[1]] + [spl1[1]]
            self.funcs_rev = True
        elif len(spl2) > 1:
            data = [data[0]] + spl2
        else:
            data.append("key")

        if data[1] == re.search('\s*', data[1], re.S | re.M).group(0):  # @UndefinedVariable
            raise IndexCreatorFunctionException("Empty function body ",
                                                len(re.split('\n', data[0])) + (len(re.split('\n', data[2])) if self.funcs_rev else 1) - 1)
        if data[2] == re.search('\s*', data[2], re.S | re.M).group(0):  # @UndefinedVariable
            raise IndexCreatorFunctionException("Empty function body ",
                                                len(re.split('\n', data[0])) + (1 if self.funcs_rev else len(re.split('\n', data[1]))) - 1)
        if data[0] == re.search('\s*', data[0], re.S | re.M).group(0):  # @UndefinedVariable
            raise IndexCreatorValueException("You didn't set any properity or you set them not at the begining of the code\n")

        data = [re.split(
            '\n', data[0]), re.split('\n', data[1]), re.split('\n', data[2])]
        self.cnt_lines = (len(data[0]), len(data[1]), len(data[2]))
        ind = 0
        self.predata = data
        self.data = [[], [], []]
        for i, v in enumerate(self.predata[0]):
            for k, w in enumerate(self.predata[0][i]):
                if self.predata[0][i][k] in self.props_assign:
                    if not is_num(self.predata[0][i][k + 1:]) and self.predata[0][i].strip()[:4] != 'type' and self.predata[0][i].strip()[:4] != 'name':
                        s = self.predata[0][i][k + 1:]
                        self.predata[0][i] = self.predata[0][i][:k + 1]

                        m = re.search('\s+', s.strip())
                        if not is_string(s) and not m:
                            s = "'" + s.strip() + "'"
                        self.predata[0][i] += s
                        break

        for n, i in enumerate(self.predata):
            for k in i:
                k = k.strip()
                if k:
                    self.data[ind].append(k)
                    self.check_enclosures(k, n)
            ind += 1

        return self.parse_ex()

    def readline(self, stage):
        def foo():
            if len(self.data[stage]) <= self.ind:
                self.ind = 0
                return ""
            else:
                self.ind += 1
                return self.data[stage][self.ind - 1]
        return foo

    def add(self, l, i):
        def add_aux(*args):
            # print args,self.ind
            if len(l[i]) < self.ind:
                l[i].append([])
            l[i][self.ind - 1].append(args)
        return add_aux

    def parse_ex(self):
        self.index_name = ""
        self.index_type = ""
        self.curLine = -1
        self.con = -1
        self.brackets = -1
        self.curFunc = None
        self.colons = 0
        self.line_cons = ([], [], [])
        self.pre_tokens = ([], [], [])
        self.known_dicts_in_mkv = []
        self.prop_name = True
        self.prop_assign = False
        self.is_one_arg_enough = False
        self.funcs_stack = []
        self.last_line = [-1, -1, -1]
        self.props_set = []
        self.custom_header = set()

        self.tokens = []
        self.tokens_head = ['# %s\n' % self.name, 'class %s(' % self.name, '):\n', '    def __init__(self, *args, **kwargs):        ']

        for i in range(3):
            tokenize.tokenize(self.readline(i), self.add(self.pre_tokens, i))
            # tokenize treats some keyword not in the right way, thats why we
            # have to change some of them
            for nk, k in enumerate(self.pre_tokens[i]):
                for na, a in enumerate(k):
                    if a[0] == token.NAME and a[1] in self.logic:
                        self.pre_tokens[i][nk][
                            na] = (token.OP, a[1], a[2], a[3], a[4])

        for i in self.pre_tokens[1]:
            self.line_cons[1].append(self.check_colons(i, 1))
            self.check_adjacents(i, 1)
            if self.check_for_2nd_arg(i) == -1 and not self.is_one_arg_enough:
                raise IndexCreatorValueException("No 2nd value to return (did u forget about ',None'?", self.cnt_line_nr(i[0][4], 1))
            self.is_one_arg_enough = False

        for i in self.pre_tokens[2]:
            self.line_cons[2].append(self.check_colons(i, 2))
            self.check_adjacents(i, 2)

        for i in self.pre_tokens[0]:
            self.handle_prop_line(i)

        self.cur_brackets = 0
        self.tokens += ['\n        super(%s, self).__init__(*args, **kwargs)\n    def make_key_value(self, data):        ' % self.name]

        for i in self.pre_tokens[1]:
            for k in i:
                self.handle_make_value(*k)

        self.curLine = -1
        self.con = -1
        self.cur_brackets = 0
        self.tokens += ['\n    def make_key(self, key):']

        for i in self.pre_tokens[2]:
            for k in i:
                self.handle_make_key(*k)

        if self.index_type == "":
            raise IndexCreatorValueException("Missing index type definition\n")
        if self.index_name == "":
            raise IndexCreatorValueException("Missing index name\n")

        self.tokens_head[0] = "# " + self.index_name + "\n" + \
            self.tokens_head[0]

        for i in self.funcs_with_body:
            if self.funcs_with_body[i][1]:
                self.tokens_head.insert(4, self.funcs_with_body[i][0])

        if None in self.custom_header:
            self.custom_header.remove(None)
        if self.custom_header:
            s = '    custom_header = """'
            for i in self.custom_header:
                s += i
            s += '"""\n'
            self.tokens_head.insert(4, s)

        if self.index_type in self.allowed_props:
            for i in self.props_set:
                if i not in self.allowed_props[self.index_type]:
                    raise IndexCreatorValueException("Properity %s is not allowed for index type: %s" % (i, self.index_type))

        # print "".join(self.tokens_head)
        # print "----------"
        # print (" ".join(self.tokens))
        return "".join(self.custom_header), "".join(self.tokens_head) + (" ".join(self.tokens))

    # has to be run BEFORE tokenize
    def check_enclosures(self, d, st):
        encs = []
        contr = {'(': ')', '{': '}', '[': ']', "'": "'", '"': '"'}
        ends = [')', '}', ']', "'", '"']
        for i in d:
            if len(encs) > 0 and encs[-1] in ['"', "'"]:
                if encs[-1] == i:
                    del encs[-1]
            elif i in contr:
                encs += [i]
            elif i in ends:
                if len(encs) < 1 or contr[encs[-1]] != i:
                    raise IndexCreatorValueException("Missing opening enclosure for \'%s\'" % i, self.cnt_line_nr(d, st))
                del encs[-1]

        if len(encs) > 0:
            raise IndexCreatorValueException("Missing closing enclosure for \'%s\'" % encs[0], self.cnt_line_nr(d, st))

    def check_adjacents(self, d, st):
        def std_check(d, n):
            if n == 0:
                prev = -1
            else:
                prev = d[n - 1][1] if d[n - 1][0] == token.OP else d[n - 1][0]

            cur = d[n][1] if d[n][0] == token.OP else d[n][0]

            # there always is an endmarker at the end, but this is a precaution
            if n + 2 > len(d):
                nex = -1
            else:
                nex = d[n + 1][1] if d[n + 1][0] == token.OP else d[n + 1][0]

            if prev not in self.allowed_adjacent[cur]:
                raise IndexCreatorValueException("Wrong left value of the %s" % cur, self.cnt_line_nr(line, st))

            # there is an assumption that whole data always ends with 0 marker, the idea prolly needs a rewritting to allow more whitespaces
            # between tokens, so it will be handled anyway
            elif nex not in self.allowed_adjacent[cur][prev]:
                raise IndexCreatorValueException("Wrong right value of the %s" % cur, self.cnt_line_nr(line, st))

        for n, (t, i, _, _, line) in enumerate(d):
            if t == token.NAME or t == token.STRING:
                if n + 1 < len(d) and d[n + 1][0] in [token.NAME, token.STRING]:
                    raise IndexCreatorValueException("Did you forget about an operator in between?", self.cnt_line_nr(line, st))
            elif i in self.allowed_adjacent:
                std_check(d, n)

    def check_colons(self, d, st):
        cnt = 0
        br = 0

        def check_ret_args_nr(a, s):
            c_b_cnt = 0
            s_b_cnt = 0
            n_b_cnt = 0
            comas_cnt = 0
            for _, i, _, _, line in a:

                if c_b_cnt == n_b_cnt == s_b_cnt == 0:
                    if i == ',':
                        comas_cnt += 1
                        if (s == 1 and comas_cnt > 1) or (s == 2 and comas_cnt > 0):
                            raise IndexCreatorFunctionException("Too much arguments to return", self.cnt_line_nr(line, st))
                        if s == 0 and comas_cnt > 0:
                            raise IndexCreatorValueException("A coma here doesn't make any sense", self.cnt_line_nr(line, st))

                    elif i == ':':
                            if s == 0:
                                raise IndexCreatorValueException("A colon here doesn't make any sense", self.cnt_line_nr(line, st))
                            raise IndexCreatorFunctionException("Two colons don't make any sense", self.cnt_line_nr(line, st))

                if i == '{':
                    c_b_cnt += 1
                elif i == '}':
                    c_b_cnt -= 1
                elif i == '(':
                    n_b_cnt += 1
                elif i == ')':
                    n_b_cnt -= 1
                elif i == '[':
                    s_b_cnt += 1
                elif i == ']':
                    s_b_cnt -= 1

        def check_if_empty(a):
            for i in a:
                if i not in [token.NEWLINE, token.INDENT, token.ENDMARKER]:
                    return False
            return True
        if st == 0:
            check_ret_args_nr(d, st)
            return

        for n, i in enumerate(d):
            if i[1] == ':':
                if br == 0:
                    if len(d) < n or check_if_empty(d[n + 1:]):
                        raise IndexCreatorValueException(
                            "Empty return value", self.cnt_line_nr(i[4], st))
                    elif len(d) >= n:
                        check_ret_args_nr(d[n + 1:], st)
                    return cnt
                else:
                    cnt += 1
            elif i[1] == '{':
                br += 1
            elif i[1] == '}':
                br -= 1
        check_ret_args_nr(d, st)
        return -1

    def check_for_2nd_arg(self, d):
        c_b_cnt = 0  # curly brackets counter '{}'
        s_b_cnt = 0  # square brackets counter '[]'
        n_b_cnt = 0  # normal brackets counter '()'

        def check_2nd_arg(d, ind):
            d = d[ind[0]:]
            for t, i, (n, r), _, line in d:
                if i == '{' or i is None:
                    return 0
                elif t == token.NAME:
                    self.known_dicts_in_mkv.append((i, (n, r)))
                    return 0
                elif t == token.STRING or t == token.NUMBER:
                    raise IndexCreatorValueException("Second return value of make_key_value function has to be a dictionary!", self.cnt_line_nr(line, 1))

        for ind in enumerate(d):
            t, i, _, _, _ = ind[1]
            if s_b_cnt == n_b_cnt == c_b_cnt == 0:
                if i == ',':
                    return check_2nd_arg(d, ind)
                elif (t == token.NAME and i not in self.funcs) or i == '{':
                    self.is_one_arg_enough = True

            if i == '{':
                c_b_cnt += 1
                self.is_one_arg_enough = True
            elif i == '}':
                c_b_cnt -= 1
            elif i == '(':
                n_b_cnt += 1
            elif i == ')':
                n_b_cnt -= 1
            elif i == '[':
                s_b_cnt += 1
            elif i == ']':
                s_b_cnt -= 1
        return -1

    def cnt_line_nr(self, l, stage):
        nr = -1
        for n, i in enumerate(self.predata[stage]):
            # print i,"|||",i.strip(),"|||",l
            if l == i.strip():
                nr = n
        if nr == -1:
            return -1

        if stage == 0:
            return nr + 1
        elif stage == 1:
            return nr + self.cnt_lines[0] + (self.cnt_lines[2] - 1 if self.funcs_rev else 0)
        elif stage == 2:
            return nr + self.cnt_lines[0] + (self.cnt_lines[1] - 1 if not self.funcs_rev else 0)

        return -1

    def handle_prop_line(self, d):
        d_len = len(d)
        if d[d_len - 1][0] == token.ENDMARKER:
            d_len -= 1

        if d_len < 3:
            raise IndexCreatorValueException("Can't handle properity assingment ", self.cnt_line_nr(d[0][4], 0))

        if not d[1][1] in self.props_assign:
            raise IndexCreatorValueException(
                "Did you forget : or =?", self.cnt_line_nr(d[0][4], 0))

        if d[0][0] == token.NAME or d[0][0] == token.STRING:
            if d[0][1] in self.props_set:
                raise IndexCreatorValueException("Properity %s is set more than once" % d[0][1], self.cnt_line_nr(d[0][4], 0))
            self.props_set += [d[0][1]]
            if d[0][1] == "type" or d[0][1] == "name":
                t, tk, _, _, line = d[2]

                if d_len > 3:
                    raise IndexCreatorValueException(
                        "Wrong value to assign", self.cnt_line_nr(line, 0))

                if t == token.STRING:
                    m = re.search('\s*(?P<a>[\'\"]+)(.*?)(?P=a)\s*', tk)
                    if m:
                        tk = m.groups()[1]
                elif t != token.NAME:
                    raise IndexCreatorValueException(
                        "Wrong value to assign", self.cnt_line_nr(line, 0))

                if d[0][1] == "type":
                    if d[2][1] == "TreeBasedIndex":
                        self.custom_header.add("from CodernityDB.tree_index import TreeBasedIndex\n")
                    elif d[2][1] == "MultiTreeBasedIndex":
                        self.custom_header.add("from CodernityDB.tree_index import MultiTreeBasedIndex\n")
                    elif d[2][1] == "MultiHashIndex":
                        self.custom_header.add("from CodernityDB.hash_index import MultiHashIndex\n")
                    self.tokens_head.insert(2, tk)
                    self.index_type = tk
                else:
                    self.index_name = tk
                return
            else:
                self.tokens += ['\n        kwargs["' + d[0][1] + '"]']
        else:
            raise IndexCreatorValueException("Can't handle properity assingment ", self.cnt_line_nr(d[0][4], 0))

        self.tokens += ['=']

        self.check_adjacents(d[2:], 0)
        self.check_colons(d[2:], 0)

        for i in d[2:]:
            self.tokens += [i[1]]

    def generate_func(self, t, tk, pos_start, pos_end, line, hdata, stage):
        if self.last_line[stage] != -1 and pos_start[0] > self.last_line[stage] and line != '':
            raise IndexCreatorFunctionException("This line will never be executed!", self.cnt_line_nr(line, stage))
        if t == 0:
            return

        if pos_start[1] == 0:
            if self.line_cons[stage][pos_start[0] - 1] == -1:
                self.tokens += ['\n        return']
                self.last_line[stage] = pos_start[0]
            else:
                self.tokens += ['\n        if']
        elif tk == ':' and self.line_cons[stage][pos_start[0] - 1] > -1:
            if self.line_cons[stage][pos_start[0] - 1] == 0:
                self.tokens += [':\n            return']
                return
            self.line_cons[stage][pos_start[0] - 1] -= 1

        if tk in self.logic2:
            # print tk
            if line[pos_start[1] - 1] != tk and line[pos_start[1] + 1] != tk:
                self.tokens += [tk]
            if line[pos_start[1] - 1] != tk and line[pos_start[1] + 1] == tk:
                if tk == '&':
                    self.tokens += ['and']
                else:
                    self.tokens += ['or']
            return

        if self.brackets != 0:
            def search_through_known_dicts(a):
                for i, (n, r) in self.known_dicts_in_mkv:
                    if i == tk and r > pos_start[1] and n == pos_start[0] and hdata == 'data':
                        return True
                return False

            if t == token.NAME and len(self.funcs_stack) > 0 and self.funcs_stack[-1][0] == 'md5' and search_through_known_dicts(tk):
                raise IndexCreatorValueException("Second value returned by make_key_value for sure isn't a dictionary ", self.cnt_line_nr(line, 1))

        if tk == ')':
            self.cur_brackets -= 1
            if len(self.funcs_stack) > 0 and self.cur_brackets == self.funcs_stack[-1][1]:
                self.tokens += [tk]
                self.tokens += self.funcs[self.funcs_stack[-1][0]][1]
                del self.funcs_stack[-1]
                return
        if tk == '(':
            self.cur_brackets += 1

        if tk in self.none:
            self.tokens += ['None']
            return

        if t == token.NAME and tk not in self.logic and tk != hdata:
            if tk not in self.funcs:
                self.tokens += [hdata + '["' + tk + '"]']
            else:
                self.tokens += self.funcs[tk][0]
                if tk in self.funcs_with_body:
                    self.funcs_with_body[tk] = (
                        self.funcs_with_body[tk][0], True)
                self.custom_header.add(self.handle_int_imports.get(tk))
                self.funcs_stack += [(tk, self.cur_brackets)]
        else:
            self.tokens += [tk]

    def handle_make_value(self, t, tk, pos_start, pos_end, line):
        self.generate_func(t, tk, pos_start, pos_end, line, 'data', 1)

    def handle_make_key(self, t, tk, pos_start, pos_end, line):
        self.generate_func(t, tk, pos_start, pos_end, line, 'key', 2)



#!/usr/bin/env python
# ppauto.py
#
# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2009, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, Auto-Discovery Service.

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""

from __future__ import absolute_import
import six.moves._thread

copyright = 'Copyright (c) 2005-2009 Vitalii Vanovschi. All rights reserved'
version = '1.5.7'

import socket
import sys
import time
import logging

# broadcast every 10 sec
BROADCAST_INTERVAL = 10


class Discover(object):
    """
    Auto-discovery service class.
    """
    def __init__(self, base, isclient=False):
        self.base = base
        self.hosts = []
        self.isclient = isclient

    def run(self, interface_addr, broadcast_addr):
        """
        Starts auto-discovery.
        """
        self.interface_addr = interface_addr
        self.broadcast_addr = broadcast_addr
        self.bsocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.bsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.bsocket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

        try:
            self.listen()
        except:
            sys.excepthook(*sys.exc_info())

    def broadcast(self):
        """
        Sends a broadcast.
        """
        if self.isclient:
            logging.debug('Client sends initial broadcast to (%s, %i)' % self.broadcast_addr)
            self.bsocket.sendto('C', self.broadcast_addr)
        else:
            while True:
                logging.debug('Server sends broadcast to (%s, %i)' % self.broadcast_addr)
                self.bsocket.sendto('S', self.broadcast_addr)
                time.sleep(BROADCAST_INTERVAL)

    def listen(self):
        """
        Listens for broadcasts from other clients/servers.
        """
        logging.debug('Listening (%s, %i)' % self.interface_addr)
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        s.bind(self.interface_addr)

        six.moves._thread.start_new_thread(self.broadcast, ())

        while True:
            try:
                message, (host, port) = s.recvfrom(1024)
                remote_address = (host, self.broadcast_addr[1])
                hostid = host + ':' + six.text_type(self.broadcast_addr[1])
                logging.debug('Discovered host (%s, %i) message=%c' % (remote_address + (message[0], )))
                if not self.base.autopp_list.get(hostid, 0) and self.isclient \
                        and message[0] == 'S':
                    logging.debug('Connecting to host %s' % (hostid, ))
                    six.moves._thread.start_new_thread(self.base.connect1, remote_address + (False, ))
                if not self.isclient and message[0] == 'C':
                    logging.debug('Replying to host %s' % (hostid, ))
                    self.bsocket.sendto('S', self.broadcast_addr)
            except:
                logging.error('An error has occured during execution of '
                              'Discover.listen', )
                sys.excepthook(*sys.exc_info())
